# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.utils.shard_utils."""
from absl.testing import parameterized
from tensorflow_datasets import testing
from tensorflow_datasets.core.utils import shard_utils


class ShardConfigTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('imagenet train, 137 GiB', 137 << 30, 1281167, True, 1024),
      ('imagenet evaluation, 6.3 GiB', 6300 * (1 << 20), 50000, True, 64),
      ('very large, but few examples, 52 GiB', 52 << 30, 512, True, 512),
      ('xxl, 10 TiB', 10 << 40, 10**9, True, 11264),
      ('xxl, 10 PiB, 100B examples', 10 << 50, 10**11, True, 10487808),
      ('xs, 100 MiB, 100K records', 10 << 20, 100 * 10**3, True, 1),
      ('m, 499 MiB, 200K examples', 400 << 20, 200 * 10**3, True, 4),
  )
  def test_get_number_shards_default_config(
      self, total_size, num_examples, uses_precise_sharding, expected_num_shards
  ):
    shard_config = shard_utils.ShardConfig()
    self.assertEqual(
        expected_num_shards,
        shard_config.get_number_shards(
            total_size=total_size,
            num_examples=num_examples,
            uses_precise_sharding=uses_precise_sharding,
        ),
    )

  def test_get_number_shards_if_specified(self):
    shard_config = shard_utils.ShardConfig(num_shards=42)
    self.assertEqual(
        42,
        shard_config.get_number_shards(
            total_size=100, num_examples=1, uses_precise_sharding=True
        ),
    )


class GetReadInstructionsTest(testing.TestCase, parameterized.TestCase):

  def test_read_all_even_sharding(self):
    # Even sharding
    res = shard_utils.get_file_instructions(
        0, 12, ['f1', 'f2', 'f3'], [4, 4, 4]
    )
    self.assertEqual(
        res,
        [
            shard_utils.FileInstruction(
                filename='f1', skip=0, take=-1, examples_in_shard=4
            ),
            shard_utils.FileInstruction(
                filename='f2', skip=0, take=-1, examples_in_shard=4
            ),
            shard_utils.FileInstruction(
                filename='f3', skip=0, take=-1, examples_in_shard=4
            ),
        ],
    )

  def test_read_all_empty_shard(self):
    res = shard_utils.get_file_instructions(
        0, 12, ['f1', 'f2', 'f3', 'f4'], [4, 4, 0, 4]
    )
    self.assertEqual(
        res,
        [
            shard_utils.FileInstruction(
                filename='f1', skip=0, take=-1, examples_in_shard=4
            ),
            shard_utils.FileInstruction(
                filename='f2', skip=0, take=-1, examples_in_shard=4
            ),
            shard_utils.FileInstruction(
                filename='f4', skip=0, take=-1, examples_in_shard=4
            ),
        ],
    )

  def test_from1_to10(self):
    res = shard_utils.get_file_instructions(
        1, 10, ['f1', 'f2', 'f3', 'f4'], [4, 4, 0, 4]
    )
    self.assertEqual(
        res,
        [
            shard_utils.FileInstruction(
                filename='f1', skip=1, take=3, examples_in_shard=4
            ),
            shard_utils.FileInstruction(
                filename='f2', skip=0, take=4, examples_in_shard=4
            ),
            shard_utils.FileInstruction(
                filename='f4', skip=0, take=2, examples_in_shard=4
            ),
        ],
    )

  def test_nothing_to_read(self):
    res = shard_utils.get_file_instructions(
        0, 0, ['f1', 'f2', 'f3', 'f4'], [0, 3, 0, 2]
    )
    self.assertEqual(res, [])
    res = shard_utils.get_file_instructions(
        4, 4, ['f1', 'f2', 'f3', 'f4'], [0, 3, 0, 2]
    )
    self.assertEqual(res, [])
    res = shard_utils.get_file_instructions(
        5, 5, ['f1', 'f2', 'f3', 'f4'], [0, 3, 0, 2]
    )
    self.assertEqual(res, [])

  def test_split_file_instruction(self):
    filename = 'data.tfrecord'
    file_instruction = shard_utils.FileInstruction(
        filename=filename, skip=0, take=-1, examples_in_shard=10
    )
    actual_splits = shard_utils.split_file_instruction(
        file_instruction=file_instruction, num_splits=3
    )
    self.assertEqual(
        actual_splits,
        [
            shard_utils.FileInstruction(
                filename=filename, skip=0, take=4, examples_in_shard=10
            ),
            shard_utils.FileInstruction(
                filename=filename, skip=4, take=4, examples_in_shard=10
            ),
            shard_utils.FileInstruction(
                filename=filename, skip=8, take=2, examples_in_shard=10
            ),
        ],
    )

  def test_split_file_instruction_one_example_many_splits(self):
    # Test when more splits are requested than there are examples.
    filename = 'data.tfrecord'
    file_instruction = shard_utils.FileInstruction(
        filename=filename, skip=0, take=-1, examples_in_shard=1
    )
    actual_splits = shard_utils.split_file_instruction(
        file_instruction=file_instruction, num_splits=9999
    )
    self.assertEqual(
        actual_splits,
        [
            shard_utils.FileInstruction(
                filename=filename, skip=0, take=1, examples_in_shard=1
            ),
        ],
    )

  @parameterized.parameters(
      ('/a/b', '/a'),
      ('a', ''),
      ('', ''),
      ('/a/b/c/d', '/a/b/c'),
  )
  def test_file_instruction_dirname(self, filename, expected):
    file_instruction = shard_utils.FileInstruction(filename, 0, -1, 1)
    self.assertEqual(file_instruction.dirname(), expected)

  @parameterized.parameters(
      ('/a/b', 'b'),
      ('a', 'a'),
      ('', ''),
      ('/a/b/c/d', 'd'),
  )
  def test_file_instruction_basename(self, filename, expected):
    file_instruction = shard_utils.FileInstruction(filename, 0, -1, 1)
    self.assertEqual(file_instruction.basename(), expected)


class FileInstructionTest(testing.TestCase, parameterized.TestCase):

  @parameterized.parameters(
      ('a', 0, 1, 1, shard_utils.FileInstruction('a', 0, 1, 1)),
      ('a', 0, 1, 2, shard_utils.FileInstruction('a', 0, 1, 2)),
      ('a', 0, -1, 1, shard_utils.FileInstruction('a', 0, 1, 1)),
      ('a', 1, -1, 10, shard_utils.FileInstruction('a', 1, 9, 10)),
  )
  def test_init(self, filename, skip, take, examples_in_shard, expected):
    actual = shard_utils.FileInstruction(
        filename, skip, take, examples_in_shard
    )
    self.assertEqual(expected, actual)

  @parameterized.parameters(
      ('a', 0, 1, -42, 'examples_in_shard should be >= 0!.+'),
      ('a', -42, 1, 2, 'skip should be between 0 and.+'),
      ('a', 0, 2, 1, 'skip .+take .+should be <= examples_in_shard.+'),
  )
  def test_incorrect_values(self, filename, skip, take, examples_in_shard, msg):
    with self.assertRaisesRegex(ValueError, expected_regex=msg):
      shard_utils.FileInstruction(filename, skip, take, examples_in_shard)


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""INaturalist2018 dataset."""

import json
import os
import urllib

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """\
There are a total of 8,142 species in the dataset, with 437,513 training images,
and 24,426 validation images. Each image has one ground truth label.
"""
_CITATION = r"""\
@misc{inaturalist18,
    Howpublished = {~\url{https://github.com/visipedia/inat_comp/tree/master/2018}},
    Title = {{iNaturalist} 2018 competition dataset.},
    Year = {2018},
    key = {{iNaturalist} 2018 competition dataset},
    }
"""
_URL = "https://ml-inat-competition-datasets.s3.amazonaws.com/2018/"


class INaturalist2018(tfds.core.GeneratorBasedBuilder):
  """Dataset from the INaturalist Competition 2018."""

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
  }

  def _info(self):
    """Define the dataset info."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "id": tfds.features.Text(),
            "image": tfds.features.Image(),
            "label": tfds.features.ClassLabel(
                names_file=tfds.core.tfds_path(
                    os.path.join(
                        "image_classification",
                        "i_naturalist2018",
                        "inaturalist2018_labels.txt",
                    )
                )
            ),
            "supercategory": tfds.features.ClassLabel(
                names_file=tfds.core.tfds_path(
                    os.path.join(
                        "image_classification",
                        "i_naturalist2018",
                        "inaturalist2018_supercategories.txt",
                    )
                )
            ),
        }),
        supervised_keys=("image", "label"),
        homepage="https://github.com/visipedia/inat_comp/tree/master/2018",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    output_files = dl_manager.download_and_extract({
        "trainval_images": tfds.download.Resource(
            url=urllib.parse.urljoin(_URL, "train_val2018.tar.gz"),
            extract_method=tfds.download.ExtractMethod.NO_EXTRACT,
        ),
        "train_annos": urllib.parse.urljoin(_URL, "train2018.json.tar.gz"),
        "val_annos": urllib.parse.urljoin(_URL, "val2018.json.tar.gz"),
        "test_images": tfds.download.Resource(
            url=urllib.parse.urljoin(_URL, "test2018.tar.gz"),
            extract_method=tfds.download.ExtractMethod.NO_EXTRACT,
        ),
        "categories": urllib.parse.urljoin(_URL, "categories.json.tar.gz"),
    })
    return {
        "train": self._generate_examples(
            images_archive=dl_manager.iter_archive(
                output_files["trainval_images"]
            ),
            annon_file=os.path.join(
                output_files["train_annos"], "train2018.json"
            ),
            category_file=os.path.join(
                output_files["categories"], "categories.json"
            ),
        ),
        "validation": self._generate_examples(
            images_archive=dl_manager.iter_archive(
                output_files["trainval_images"]
            ),
            annon_file=os.path.join(output_files["val_annos"], "val2018.json"),
            category_file=os.path.join(
                output_files["categories"], "categories.json"
            ),
        ),
        "test": self._generate_examples(
            images_archive=dl_manager.iter_archive(output_files["test_images"]),
            annon_file=None,
            category_file=None,
        ),
    }

  def _generate_examples(self, images_archive, annon_file, category_file):
    """Generate examples."""
    if annon_file is not None:
      # Training and validation images.
      with tf.io.gfile.GFile(annon_file, "r") as f:
        data = json.load(f)
      with tf.io.gfile.GFile(category_file, "r") as f:
        categories = json.load(f)

        # First read the annotations file, used to filter the contents of the
      # tar.gz file when yielding examples.
      key2data = {}
      for image, annotation in zip(data["images"], data["annotations"]):
        category_id = annotation["category_id"]
        category = categories[category_id]["name"]
        supercategory = data["categories"][category_id]["supercategory"]
        key = os.path.basename(image["file_name"]).split(".")[0]
        key2data[key] = {
            "id": key,
            "label": category,
            "supercategory": supercategory,
        }
      # Read tar.gz file containing train & validation images and yield relevant
      # examples.
      for fpath, fobj in images_archive:
        key = os.path.basename(fpath).split(".")[0]
        if key in key2data:
          data = key2data[key].copy()
          data["image"] = fobj
          yield key, data
    else:
      # Read tar.gz file containing all test images and yield all examples.
      for fpath, fobj in images_archive:
        key = os.path.basename(fpath).split(".")[0]
        # Note: test labels are not annotated, so just return -1 as labels.
        yield key, {
            "id": key,
            "image": fobj,
            "label": -1,
            "supercategory": -1,
        }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for compute_split_info."""
import pytest

from tensorflow_datasets import testing
from tensorflow_datasets.core import naming
from tensorflow_datasets.core.folder_dataset import compute_split_utils


def test_compute_split_info(tmp_path):
  builder = testing.DummyDataset(data_dir=tmp_path)
  builder.download_and_prepare()

  filename_template = naming.ShardedFileTemplate(
      dataset_name=builder.name,
      data_dir=builder.data_dir,
      filetype_suffix=builder.info.file_format.file_suffix,
  )
  split_infos = compute_split_utils.compute_split_info(
      out_dir=tmp_path,
      filename_template=filename_template,
  )

  assert [s.to_proto() for s in split_infos] == [
      s.to_proto() for s in builder.info.splits.values()
  ]

  # Split info are correctly saved
  filename_template = filename_template.replace(
      data_dir=tmp_path, split='train'
  )
  split_info = compute_split_utils._split_info_from_path(filename_template)
  assert builder.info.splits['train'].to_proto() == split_info.to_proto()


def test_enrich_filename_template():
  template = naming.ShardedFileTemplate(data_dir='/path')
  files_per_split = {
      'train': [
          naming.FilenameInfo(
              dataset_name='mnist',
              filetype_suffix='tfrecord',
              split=None,
              shard_index=None,
              num_shards=None,
          )
      ],
  }
  actual = compute_split_utils._enrich_filename_template(
      filename_template=template, files_per_split=files_per_split
  )
  assert actual.dataset_name == 'mnist'
  assert actual.filetype_suffix == 'tfrecord'


def test_enrich_filename_template_multiple_dataset_names():
  template = naming.ShardedFileTemplate(data_dir='/path')
  files_per_split = {
      'train': [
          naming.FilenameInfo(
              dataset_name='a',
              filetype_suffix='tfrecord',
              split=None,
              shard_index=None,
              num_shards=None,
          ),
          naming.FilenameInfo(
              dataset_name='b',
              filetype_suffix='tfrecord',
              split=None,
              shard_index=None,
              num_shards=None,
          ),
      ],
  }
  with pytest.raises(ValueError, match='Found multiple dataset names'):
    compute_split_utils._enrich_filename_template(
        filename_template=template, files_per_split=files_per_split
    )


def test_enrich_filename_template_multiple_filetype_suffixes():
  template = naming.ShardedFileTemplate(data_dir='/path')
  files_per_split = {
      'train': [
          naming.FilenameInfo(
              dataset_name='mnist',
              filetype_suffix='a',
              split=None,
              shard_index=None,
              num_shards=None,
          ),
          naming.FilenameInfo(
              dataset_name='mnist',
              filetype_suffix='b',
              split=None,
              shard_index=None,
              num_shards=None,
          ),
      ],
  }
  with pytest.raises(ValueError, match='Found multiple filetype suffixes'):
    compute_split_utils._enrich_filename_template(
        filename_template=template, files_per_split=files_per_split
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configures tests for C4-WSRS dataset."""
import contextlib
from unittest import mock

import tensorflow as tf
from tensorflow_datasets import testing
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text.c4_wsrs import c4_wsrs


class C4WSRSTest(testing.DatasetBuilderTestCase):
  """Tests for c4_wsrs dataset."""

  DATASET_CLASS = c4_wsrs.C4WSRS
  SPLITS = {
      'train': 2,  # Number of fake train example
      'validation': 2,  # Number of fake test example
  }

  OVERLAPPING_SPLITS = ['train', 'validation']
  SKIP_CHECKSUMS = True
  DL_DOWNLOAD_RESULT = 'abbreviation_expansion_dictionary.csv'
  SKIP_TF1_GRAPH_MODE = True

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    beam = tfds.core.lazy_imports.apache_beam
    inputs = [
        tf.nest.map_structure(
            tf.convert_to_tensor,
            {
                'content-length': 'abc',
                'content-type': 'text/plain',
                'text': (
                    'patient in the emergency room with complications. '
                    'will check in morning.'
                ),
                'timestamp': 'this is a timestamp',
                'url': 'http://google.com/1',
            },
        ),
        tf.nest.map_structure(
            tf.convert_to_tensor,
            {
                'content-length': 'abc',
                'content-type': 'text/plain',
                'text': (
                    'magnetic resonance imaging test in patient with '
                    'lower back pain. send for treatment.'
                ),
                'timestamp': 'this is a timestamp',
                'url': 'http://google.com/2',
            },
        ),
    ]

    cls._exit_stack = contextlib.ExitStack().__enter__()
    cls._exit_stack.enter_context(
        mock.patch.object(tfds, 'builder', autospec=True)
    )
    cls._exit_stack.enter_context(
        mock.patch.object(
            tfds.beam,
            'ReadFromTFDS',
            return_value=beam.Create(inputs),
        )
    )

  @classmethod
  def tearDownClass(cls):
    cls._exit_stack.__exit__(None, None, None)
    super().tearDownClass()


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""youtube_vis dataset."""

import numpy as np
import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.video.youtube_vis import youtube_vis


class YoutubeVisTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for youtube_vis dataset."""

  DATASET_CLASS = youtube_vis.YoutubeVis
  SPLITS = {
      tfds.Split.TRAIN: 1,  # Number of fake train examples.
      tfds.Split.VALIDATION: 1,  # Number of fake test examples.
      tfds.Split.TEST: 1,  # Number of fake test examples.
  }
  SKIP_CHECKSUMS = True  # All data is manually downloaded.

  # When testing the "custom split" functionality the splits are overlapping.
  OVERLAPPING_SPLITS = [
      tfds.Split.TRAIN,
      tfds.Split.VALIDATION,
      tfds.Split.TEST,
  ]

  BUILDER_CONFIG_NAMES_TO_TEST = ['test_config', 'test_config_custom_split']

  @classmethod
  def setUpClass(cls):
    youtube_vis.YoutubeVis.BUILDER_CONFIGS = [
        youtube_vis.YoutubeVisConfig(
            name='test_config',
            description='All images are bilinearly resized to 28 X 42',
            height=28,
            width=42,
        ),
        youtube_vis.YoutubeVisConfig(
            name='test_config_custom_split',
            description='All splits are the train data resized to 28 X 42.',
            height=28,
            width=42,
            split_train_data_range=(0, 1),
            split_val_data_range=(0, 1),
            split_test_data_range=(0, 1),
        ),
    ]
    super().setUpClass()

  def _download_and_prepare_as_dataset(self, builder):
    super()._download_and_prepare_as_dataset(builder)

    if not tf.executing_eagerly():  # Only test the following in eager mode.
      return

    with self.subTest('check_annotations'):
      # Makes certain sanity checks like the number of segmentations
      # should be equal to the number of labeled frame indices, etc.
      splits = builder.as_dataset()
      train_ex = list(splits[tfds.Split.TRAIN])[0]
      val_ex = list(splits[tfds.Split.VALIDATION])[0]
      test_ex = list(splits[tfds.Split.TEST])[0]
      self.assertEqual(
          train_ex['tracks']['bboxes'].shape[0],
          train_ex['tracks']['segmentations'].shape[0],
      )
      self.assertEqual(
          train_ex['tracks']['segmentations'].shape[0],
          train_ex['tracks']['frames'].shape[0],
      )
      if builder.builder_config.name != 'test_config_custom_split':
        # No annotations are provided on the val and test data if not using
        # the custom split.
        self.assertEqual(val_ex['tracks']['bboxes'].shape[0], 0)
        self.assertEqual(test_ex['tracks']['bboxes'].shape[0], 0)

    with self.subTest('check_video'):
      # Checks that the video data matches the metadata.
      splits = builder.as_dataset()
      train_ex = list(splits[tfds.Split.TRAIN])[0]
      val_ex = list(splits[tfds.Split.VALIDATION])[0]
      test_ex = list(splits[tfds.Split.TEST])[0]

      self.assertEqual(
          train_ex['metadata']['height'].numpy(), builder.builder_config.height
      )
      self.assertEqual(
          train_ex['metadata']['width'].numpy(), builder.builder_config.width
      )
      self.assertEqual(
          val_ex['metadata']['height'].numpy(), builder.builder_config.height
      )
      self.assertEqual(
          val_ex['metadata']['width'].numpy(), builder.builder_config.width
      )
      self.assertEqual(
          test_ex['metadata']['height'].numpy(), builder.builder_config.height
      )
      self.assertEqual(
          test_ex['metadata']['width'].numpy(), builder.builder_config.width
      )
      self.assertEqual(
          train_ex['video'].shape,
          (
              train_ex['metadata']['num_frames'].numpy(),
              train_ex['metadata']['height'].numpy(),
              train_ex['metadata']['width'].numpy(),
              3,
          ),
      )
      self.assertEqual(
          val_ex['video'].shape,
          (
              val_ex['metadata']['num_frames'].numpy(),
              val_ex['metadata']['height'].numpy(),
              val_ex['metadata']['width'].numpy(),
              3,
          ),
      )
      self.assertEqual(
          test_ex['video'].shape,
          (
              test_ex['metadata']['num_frames'].numpy(),
              test_ex['metadata']['height'].numpy(),
              test_ex['metadata']['width'].numpy(),
              3,
          ),
      )

    with self.subTest('check_segmentations'):
      # Checks that the segmentations match the metadata and have the correct
      # label count.
      splits = builder.as_dataset()
      train_ex = list(splits[tfds.Split.TRAIN])[0]
      for segmentation in train_ex['tracks']['segmentations']:
        unique_pixels = set(np.unique(segmentation.numpy()))
        self.assertSetEqual(unique_pixels, set([0, 1]))
        self.assertEqual(
            (
                train_ex['metadata']['height'].numpy(),
                train_ex['metadata']['width'].numpy(),
            ),
            segmentation.shape[1:3],
        )

    if builder.builder_config.name == 'test_config_custom_split':
      with self.subTest('check_custom_split'):
        # Checks that all splits contain exactly the same data, since
        # that is what was requested by this custom split.
        splits = builder.as_dataset()
        train_ex = list(splits[tfds.Split.TRAIN])[0]
        val_ex = list(splits[tfds.Split.VALIDATION])[0]
        val_ex = list(splits[tfds.Split.TEST])[0]
        self.assertDictEqual(train_ex['metadata'], val_ex['metadata'])
        self.assertDictEqual(test_ex['metadata'], val_ex['metadata'])


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tensorflow_datasets.image_classification import caltech_birds
import tensorflow_datasets.testing as tfds_test


class CaltechBirdsTest(tfds_test.DatasetBuilderTestCase):
  DATASET_CLASS = caltech_birds.CaltechBirds2010

  SPLITS = {  # No. of train and test samples
      'train': 9,
      'test': 6,
  }

  DL_EXTRACT_RESULT = ['Lists.tgz', 'Annotations.tgz', 'Images.tar.gz']


class CaltechBirds2011Test(tfds_test.DatasetBuilderTestCase):
  DATASET_CLASS = caltech_birds.CaltechBirds2011

  SPLITS = {  # No. of train and test samples
      'train': 6,
      'test': 4,
  }

  DL_EXTRACT_RESULT = [
      'extracted/TAR_GZ.CUB_200_2011.tar.gz',
      'extracted/segmentations.tgz',
  ]

  DL_DOWNLOAD_RESULT = ['CUB_200_2011.tar.gz']


if __name__ == '__main__':
  tfds_test.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""web_graph dataset."""

from __future__ import annotations

import dataclasses
import os
import textwrap
from typing import List

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
This dataset contains a sparse graph representing web link structure for a
small subset of the Web.

Its a processed version of a single crawl performed by CommonCrawl in 2021
where we strip everything and keep only the link->outlinks structure.
The final dataset is basically int -> List[int] format with each integer id
representing a url.

Also, in order to increase the value of this resource, we created 6 different
version of WebGraph, each varying in the sparsity pattern and locale. We took
the following processing steps, in order:

- We started with WAT files from June 2021 crawl.
- Since the outlinks in HTTP-Response-Metadata are stored as relative paths, we
convert them to absolute paths using urllib after validating each link.
- To study locale-specific graphs, we further filter based on 2 top level
domains: ‘de’ and ‘in’, each producing a graph with an order of magnitude less
number of nodes.
- These graphs can still have arbitrary sparsity patterns and dangling links.
Thus we further filter the nodes in each graph to have minimum of K ∈ [10, 50]
inlinks and outlinks. Note that we only do this processing once, thus this is
still an approximation i.e. the resulting graph might have nodes with less than
K links.
- Using both locale and count filters, we finalize 6 versions of WebGraph
dataset, summarized in the folling table.

| Version      | Top level domain | Min count | Num nodes | Num edges |
| -----------  | ---------------- | --------- | --------- | --------- |
| sparse       |                  | 10        | 365.4M    | 30B       |
| dense        |                  | 50        | 136.5M    | 22B       |
| de-sparse    | de               | 10        | 19.7M     | 1.19B     |
| de-dense     | de               | 50        | 5.7M      | 0.82B     |
| in-sparse    | in               | 10        | 1.5M      | 0.14B     |
| in-dense     | in               | 50        | 0.5M      | 0.12B     |

All versions of the dataset have following features:

- "row_tag": a unique identifier of the row (source link).
- "col_tag": a list of unique identifiers of non-zero columns (dest outlinks).
- "gt_tag": a list of unique identifiers of non-zero columns used as ground
truth (dest outlinks), empty for train/train_t splits.
"""

_CITATION = """
@article{mehta2021alx,
    title={ALX: Large Scale Matrix Factorization on TPUs},
    author={Harsh Mehta and Steffen Rendle and Walid Krichene and Li Zhang},
    year={2021},
    eprint={2112.02194},
    archivePrefix={arXiv},
    primaryClass={cs.LG}
}
"""


@dataclasses.dataclass
class WebGraphConfig(tfds.core.BuilderConfig):
  """Palmer Penguins dataset builder config."""

  # Basename of the file hosting the data.
  subfolder: str = ''
  num_shards: int = 1


def filename(subfolder: str, name: str):
  full_name = os.path.join(subfolder, name)
  return os.fspath(tfds.core.Path(full_name))


class WebGraph(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for web_graph dataset."""

  VERSION = tfds.core.Version('1.0.0')
  WEB_GRAPH_HOMEPAGE = 'gs://gresearch/web_graph/'
  SHARDS = None
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = [
      WebGraphConfig(
          name='sparse',
          subfolder='web_graph_min_10',
          num_shards=1000,
          description=textwrap.dedent(
              """\
              WebGraph-sparse contains around 30B edges and around 365M nodes.
              """
          ),
      ),
      WebGraphConfig(
          name='dense',
          subfolder='web_graph_min_50',
          num_shards=1000,
          description=textwrap.dedent(
              """\
              WebGraph-dense contains around 22B edges and around 136.5M nodes.
              """
          ),
      ),
      WebGraphConfig(
          name='de-sparse',
          subfolder='web_graph_tld_de_min_10',
          num_shards=100,
          description=textwrap.dedent(
              """\
              WebGraph-de-sparse contains around 1.19B edges and around 19.7M
              nodes.
              """
          ),
      ),
      WebGraphConfig(
          name='de-dense',
          subfolder='web_graph_tld_de_min_50',
          num_shards=100,
          description=textwrap.dedent(
              """\
              WebGraph-de-dense contains around 0.82B edges and around 5.7M
              nodes.
              """
          ),
      ),
      WebGraphConfig(
          name='in-sparse',
          subfolder='web_graph_tld_in_min_10',
          num_shards=10,
          description=textwrap.dedent(
              """\
              WebGraph-de-sparse contains around 0.14B edges and around 1.5M
              nodes.
              """
          ),
      ),
      WebGraphConfig(
          name='in-dense',
          subfolder='web_graph_tld_in_min_50',
          num_shards=10,
          description=textwrap.dedent(
              """\
              WebGraph-de-dense contains around 0.12B edges and around 0.5M
              nodes.
              """
          ),
      ),
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'row_tag': np.int64,
            'col_tag': tfds.features.Sequence(np.int64),
            'gt_tag': tfds.features.Sequence(np.int64),
        }),
        # If there's a common (input, target) tuple from the
        # features, specify them here. They'll be used if
        # `as_supervised=True` in `builder.as_dataset`.
        supervised_keys=None,  # Set to `None` to disable
        homepage='https://arxiv.org/abs/2112.02194',
        citation=_CITATION,
    )

  def _split_generators(
      self, dl_manager: tfds.download.DownloadManager, pipeline
  ):
    """Returns SplitGenerators."""
    del dl_manager

    subfolder = os.path.join(
        self.WEB_GRAPH_HOMEPAGE, self.builder_config.subfolder
    )
    shards = self.SHARDS or self.builder_config.num_shards

    train_files = [
        filename(subfolder, f'train.tfr-{i:05}-of-{int(shards):05}')
        for i in range(shards)
    ]
    train_t_files = [
        filename(subfolder, f'train.transpose.tfr-{i:05}-of-{int(shards):05}')
        for i in range(shards)
    ]
    test_files = [
        filename(subfolder, f'test.tfr-{i:05}-of-{int(shards):05}')
        for i in range(shards)
    ]
    return {
        'train': self._generate_examples(pipeline, train_files, split='train'),
        'train_t': self._generate_examples(
            pipeline, train_t_files, split='train_t'
        ),
        'test': self._generate_examples(pipeline, test_files, split='test'),
    }

  def _generate_examples(self, pipeline, files, split: str):
    """Yields examples."""
    beam = tfds.core.lazy_imports.apache_beam

    def _get_int_feature(
        example: tf.train.Example, feature_name: str
    ) -> List[int]:
      return example.features.feature[feature_name].int64_list.value

    def _process_example(example: bytes, is_test=False):
      example = tf.train.Example.FromString(example)
      row_tag = _get_int_feature(example, 'row_tag')[0]
      col_tag = np.array(_get_int_feature(example, 'col_tag'), dtype=np.int64)
      if is_test:
        gt_tag = _get_int_feature(example, 'gt_tag')
      else:
        gt_tag = []
      gt_tag = np.array(gt_tag, dtype=np.int64)
      return_dict = {'row_tag': row_tag, 'col_tag': col_tag, 'gt_tag': gt_tag}
      return row_tag, return_dict

    return (
        pipeline
        | f'{split}_create' >> beam.Create(files)
        | f'{split}_read' >> beam.io.tfrecordio.ReadAllFromTFRecord()
        | f'{split}_process' >> beam.Map(_process_example, split == 'test')
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.reader."""

import functools
import itertools
import os
from unittest import mock

import tensorflow as tf
import tensorflow_datasets as tfds
from tensorflow_datasets import testing
from tensorflow_datasets.core import example_parser
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import reader as reader_lib
from tensorflow_datasets.core import splits
from tensorflow_datasets.core import writer as writer_lib
from tensorflow_datasets.core.utils import read_config as read_config_lib
from tensorflow_datasets.core.utils import shard_utils

# Skip the cardinality test for backward compatibility with TF <= 2.1.
_SKIP_CARDINALITY_TEST = not hasattr(tf.data.experimental, 'assert_cardinality')

_SHUFFLE_FILES_ERROR_MESSAGE = (
    'Dataset is an ordered dataset '
    "('disable_shuffling=True'), but examples "
    'will not be read in order because '
    '`shuffle_files=True`.'
)
_CYCLE_LENGTH_ERROR_MESSAGE = (
    'Dataset is an ordered dataset '
    "('disable_shuffling=True'), but examples will"
    ' not be read in order because '
    '`ReadConfig.interleave_cycle_length != 1`.'
)


def _write_tfrecord_from_shard_spec(shard_spec, get):
  """Write tfrecord shard given shard_spec and buckets to read data from.

  Args:
    shard_spec: _ShardSpec, the spec for shard to write.
    get: callable taking the shard index (of bucket) and returning iterator over
      its elements.
  """
  iterators = []
  for instruction in shard_spec.file_instructions:
    iterator = get(int(instruction.filename))
    skip, take = instruction.skip, instruction.take
    stop = skip + take if take > 0 else None
    iterators.append(itertools.islice(iterator, skip, stop))
  writer_lib._write_examples(shard_spec.path, itertools.chain(*iterators))


class ReaderTest(testing.TestCase):

  def setUp(self):
    super(ReaderTest, self).setUp()
    with mock.patch.object(
        example_parser, 'ExampleParser', testing.DummyParser
    ):
      self.reader = reader_lib.Reader(self.tmp_dir, 'some_spec')
      self.reader.read = functools.partial(
          self.reader.read,
          read_config=read_config_lib.ReadConfig(),
          shuffle_files=False,
      )

  def _filename_template(self, split: str) -> naming.ShardedFileTemplate:
    return naming.ShardedFileTemplate(
        dataset_name='mnist',
        split=split,
        filetype_suffix='tfrecord',
        data_dir=self.tmp_dir,
    )

  def _write_tfrecord(self, split_name, shards_number, records):
    filename_template = self._filename_template(split=split_name)
    num_examples = len(records)
    shard_specs = writer_lib._get_shard_specs(
        num_examples=num_examples,
        total_size=0,
        bucket_lengths=[num_examples],
        filename_template=filename_template,
        shard_config=shard_utils.ShardConfig(num_shards=shards_number),
    )
    serialized_records = [
        (key, bytes(rec, encoding='utf-8')) for key, rec in enumerate(records)
    ]
    for shard_spec in shard_specs:
      _write_tfrecord_from_shard_spec(
          shard_spec, lambda unused_i: iter(serialized_records)
      )
    return splits.SplitInfo(
        name=split_name,
        shard_lengths=[int(s.examples_number) for s in shard_specs],
        num_bytes=0,
        filename_template=filename_template,
    )

  def test_nodata_instruction(self):
    # Given instruction corresponds to no data.
    with self.assertRaisesWithPredicateMatch(
        ValueError, 'corresponds to no data!'
    ):
      train_info = splits.SplitInfo(
          name='train',
          shard_lengths=[2, 3, 2, 3, 2],
          num_bytes=0,
          filename_template=self._filename_template(split='train'),
      )
      self.reader.read(
          instructions='train[0:0]',
          split_infos=[train_info],
      )

  def test_noskip_notake(self):
    train_info = self._write_tfrecord('train', 5, 'abcdefghijkl')
    ds = self.reader.read(
        instructions='train',
        split_infos=[train_info],
    )
    read_data = list(tfds.as_numpy(ds))
    self.assertEqual(
        read_data, [bytes(l, encoding='utf-8') for l in 'abcdefghijkl']
    )

    if not _SKIP_CARDINALITY_TEST:
      # Check that the cardinality is correctly set.
      self.assertEqual(ds.cardinality().numpy(), len(read_data))

  def test_overlap(self):
    train_info = self._write_tfrecord('train', 5, 'abcdefghijkl')
    ds = self.reader.read(
        instructions='train+train[:2]',
        split_infos=[train_info],
    )
    read_data = list(tfds.as_numpy(ds))
    self.assertEqual(
        read_data, [bytes(l, encoding='utf-8') for l in 'abcdefghijklab']
    )

    if not _SKIP_CARDINALITY_TEST:
      # Check that the cardinality is correctly set.
      self.assertEqual(ds.cardinality().numpy(), len(read_data))

  def test_complex(self):
    train_info = self._write_tfrecord('train', 5, 'abcdefghijkl')
    test_info = self._write_tfrecord('test', 3, 'mnopqrs')
    self.assertEqual(train_info.name, 'train')
    self.assertEqual(test_info.name, 'test')
    self.assertEqual(train_info.shard_lengths, [2, 3, 2, 3, 2])  # 12 ex.
    self.assertEqual(test_info.shard_lengths, [2, 3, 2])  # 7 ex.
    split_info = [train_info, test_info]
    ds = self.reader.read(
        instructions='train[1:-1]+test[:-50%]',
        split_infos=split_info,
    )
    read_data = list(tfds.as_numpy(ds))
    self.assertEqual(
        read_data, [bytes(l, encoding='utf-8') for l in 'bcdefghijkmno']
    )

    if not _SKIP_CARDINALITY_TEST:
      # Check that the cardinality is correctly set.
      self.assertEqual(ds.cardinality().numpy(), len(read_data))

  def test_shuffle_files(self):
    chars = 'abcdefghijkl'
    train_info = self._write_tfrecord('train', 5, chars)
    ds = self.reader.read(
        instructions='train',
        split_infos=[train_info],
        shuffle_files=True,
    )
    ds = ds.batch(12).repeat(100)
    read_data = set(tuple(e) for e in tfds.as_numpy(ds))
    for batch in read_data:
      # Check that `batch` contains all the chars exactly once.
      batch_set = set(batch)
      self.assertEqual(len(batch_set), len(chars))
    # There are theoretically 5! (=120) different arrangements, but we would
    # need too many repeats to be sure to get them.
    self.assertGreater(len(set(read_data)), 10)

  def test_shuffle_deterministic(self):
    split_info = self._write_tfrecord('train', 5, 'abcdefghijkl')
    read_config = read_config_lib.ReadConfig(
        shuffle_seed=123,
    )
    ds = self.reader.read(
        instructions='train',
        split_infos=[split_info],
        read_config=read_config,
        shuffle_files=True,
    )
    ds_values = list(tfds.as_numpy(ds))

    # Check that shuffle=True with a seed provides deterministic results.
    self.assertEqual(
        ds_values,
        [
            b'a',
            b'b',
            b'k',
            b'l',
            b'h',
            b'i',
            b'j',
            b'c',
            b'd',
            b'e',
            b'f',
            b'g',
        ],
    )

  def test_4fold(self):
    train_info = self._write_tfrecord('train', 5, 'abcdefghijkl')
    instructions = [
        splits.ReadInstruction('train', from_=k, to=k + 25, unit='%')
        for k in range(0, 100, 25)
    ]
    tests = self.reader.read(
        instructions=instructions,
        split_infos=[train_info],
    )
    instructions = [
        (
            splits.ReadInstruction('train', to=k, unit='%')
            + splits.ReadInstruction('train', from_=k + 25, unit='%')
        )
        for k in range(0, 100, 25)
    ]
    trains = self.reader.read(
        instructions=instructions,
        split_infos=[train_info],
    )
    read_tests = [list(r) for r in tfds.as_numpy(tests)]
    read_trains = [list(r) for r in tfds.as_numpy(trains)]
    self.assertEqual(
        read_tests,
        [
            [b'a', b'b', b'c'],
            [b'd', b'e', b'f'],
            [b'g', b'h', b'i'],
            [b'j', b'k', b'l'],
        ],
    )
    self.assertEqual(
        read_trains,
        [
            [b'd', b'e', b'f', b'g', b'h', b'i', b'j', b'k', b'l'],
            [b'a', b'b', b'c', b'g', b'h', b'i', b'j', b'k', b'l'],
            [b'a', b'b', b'c', b'd', b'e', b'f', b'j', b'k', b'l'],
            [b'a', b'b', b'c', b'd', b'e', b'f', b'g', b'h', b'i'],
        ],
    )

  def test_read_files(self):
    self._write_tfrecord('train', 4, 'abcdefghijkl')
    filename_template = self._filename_template(split='train')
    ds = self.reader.read_files(
        [
            shard_utils.FileInstruction(
                filename=os.fspath(
                    filename_template.sharded_filepath(
                        shard_index=1, num_shards=4
                    )
                ),
                skip=0,
                take=-1,
                examples_in_shard=3,
            ),
            shard_utils.FileInstruction(
                filename=os.fspath(
                    filename_template.sharded_filepath(
                        shard_index=3, num_shards=4
                    )
                ),
                skip=1,
                take=1,
                examples_in_shard=3,
            ),
        ],
        read_config=read_config_lib.ReadConfig(),
        shuffle_files=False,
    )
    read_data = list(tfds.as_numpy(ds))
    self.assertEqual(read_data, [bytes(l, encoding='utf-8') for l in 'defk'])

  def test_input_context(self):
    split_info = self._write_tfrecord('train', 5, 'abcdefghijkl')
    self.assertEqual(split_info.shard_lengths, [2, 3, 2, 3, 2])

    def read(num_workers, index):
      return list(
          tfds.as_numpy(
              self.reader.read(
                  instructions='train',
                  split_infos=[split_info],
                  read_config=read_config_lib.ReadConfig(
                      input_context=tf.distribute.InputContext(
                          num_input_pipelines=num_workers,
                          input_pipeline_id=index,
                      ),
                  ),
                  # Workers should read a deterministic subset of the examples,
                  # even if examples within one worker may be shuffled.
                  shuffle_files=True,
              )
          )
      )

    def _b(bytes_str):
      # Convert to List[bytes] (rather than List[int])
      return [bytes([b]) for b in bytes_str]

    # Read all the data (single pipeline)
    self.assertCountEqual(read(num_workers=1, index=0), _b(b'abcdefghijkl'))
    # Read part of the data (workers should not overlapp)
    self.assertCountEqual(read(num_workers=3, index=0), _b(b'abhij'))  # 0, 3
    self.assertCountEqual(read(num_workers=3, index=1), _b(b'cdekl'))  # 1, 4
    self.assertEqual(read(num_workers=3, index=2), _b(b'fg'))  # Shards 2
    # If num_workers == num_shards, then a single shard is read
    self.assertEqual(read(num_workers=5, index=1), _b(b'cde'))  # Shard 1
    # If num_workers > num_shards, raise error
    with self.assertRaisesRegex(ValueError, 'Cannot shard the pipeline'):
      read(num_workers=6, index=0)

  def test_shuffle_files_should_be_disabled(self):
    self._write_tfrecord('train', 4, 'abcdefghijkl')
    filename_template = self._filename_template(split='train')
    with self.assertRaisesWithPredicateMatch(
        ValueError, _SHUFFLE_FILES_ERROR_MESSAGE
    ):
      self.reader.read_files(
          [
              shard_utils.FileInstruction(
                  filename=os.fspath(
                      filename_template.sharded_filepath(
                          shard_index=1, num_shards=4
                      )
                  ),
                  skip=0,
                  take=-1,
                  examples_in_shard=3,
              ),
          ],
          read_config=read_config_lib.ReadConfig(),
          shuffle_files=True,
          disable_shuffling=True,
      )

  def test_cycle_length_must_be_one(self):
    self._write_tfrecord('train', 4, 'abcdefghijkl')
    filename_template = self._filename_template(split='train')
    instructions = [
        shard_utils.FileInstruction(
            filename=os.fspath(
                filename_template.sharded_filepath(shard_index=1, num_shards=4)
            ),
            skip=0,
            take=-1,
            examples_in_shard=3,
        ),
    ]
    # In ordered dataset interleave_cycle_length is set to 1 by default
    self.reader.read_files(
        instructions,
        read_config=read_config_lib.ReadConfig(),
        shuffle_files=False,
        disable_shuffling=True,
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, _CYCLE_LENGTH_ERROR_MESSAGE
    ):
      self.reader.read_files(
          instructions,
          read_config=read_config_lib.ReadConfig(interleave_cycle_length=16),
          shuffle_files=False,
          disable_shuffling=True,
      )

  def test_ordering_guard(self):
    self._write_tfrecord('train', 4, 'abcdefghijkl')
    filename_template = self._filename_template(split='train')
    instructions = [
        shard_utils.FileInstruction(
            filename=filename_template.sharded_filepath(
                shard_index=1, num_shards=4
            ),
            skip=0,
            take=-1,
            examples_in_shard=3,
        ),
    ]
    reported_warnings = []
    with mock.patch('absl.logging.warning', reported_warnings.append):
      self.reader.read_files(
          instructions,
          read_config=read_config_lib.ReadConfig(
              interleave_cycle_length=16, enable_ordering_guard=False
          ),
          shuffle_files=True,
          disable_shuffling=True,
      )
      expected_warning = (
          _SHUFFLE_FILES_ERROR_MESSAGE + '\n' + _CYCLE_LENGTH_ERROR_MESSAGE
      )
      self.assertIn(expected_warning, reported_warnings)

  @mock.patch(
      'tensorflow.data.experimental.assert_cardinality',
      wraps=tf.data.experimental.assert_cardinality,
  )
  def test_assert_cardinality_is_on_by_default(self, assert_cardinality):
    train_info = self._write_tfrecord('train', 5, 'abcdefghijkl')
    self.reader.read(instructions='train', split_infos=[train_info])
    assert_cardinality.assert_called_with(12)

  @mock.patch('tensorflow.data.experimental.assert_cardinality')
  def test_assert_cardinality_can_be_disabled_through_readconfig(
      self, assert_cardinality
  ):
    train_info = self._write_tfrecord('train', 5, 'abcdefghijkl')
    self.reader.read(
        instructions='train',
        split_infos=[train_info],
        read_config=read_config_lib.ReadConfig(assert_cardinality=False),
    )
    assert not assert_cardinality.called


def test_shard_api():
  si = tfds.core.SplitInfo(
      name='train',
      shard_lengths=[10, 20, 13],
      num_bytes=0,
      filename_template=naming.ShardedFileTemplate(
          dataset_name='ds_name',
          split='train',
          filetype_suffix='tfrecord',
          data_dir='/path',
      ),
  )
  fi = [
      shard_utils.FileInstruction(
          filename='/path/ds_name-train.tfrecord-00000-of-00003',
          skip=0,
          take=-1,
          examples_in_shard=10,
      ),
      shard_utils.FileInstruction(
          filename='/path/ds_name-train.tfrecord-00001-of-00003',
          skip=0,
          take=-1,
          examples_in_shard=20,
      ),
      shard_utils.FileInstruction(
          filename='/path/ds_name-train.tfrecord-00002-of-00003',
          skip=0,
          take=-1,
          examples_in_shard=13,
      ),
  ]
  sd = splits.SplitDict([si])
  assert sd['train[0shard]'].file_instructions == [fi[0]]
  assert sd['train[1shard]'].file_instructions == [fi[1]]
  assert sd['train[-1shard]'].file_instructions == [fi[-1]]
  assert sd['train[-2shard]'].file_instructions == [fi[-2]]
  assert sd['train[:2shard]'].file_instructions == fi[:2]
  assert sd['train[1shard:]'].file_instructions == fi[1:]
  assert sd['train[-1shard:]'].file_instructions == fi[-1:]
  assert sd['train[1:-1shard]'].file_instructions == fi[1:-1]


class AddTfdsIdTest(testing.TestCase):

  def test_get_tfds_id_prefixes_single_directory(self):
    file_instructions = [
        shard_utils.FileInstruction('/a/b/f1', 0, -1, 1),
        shard_utils.FileInstruction('/a/b/f2', 0, -1, 1),
    ]
    actual = reader_lib._get_tfds_id_prefixes(file_instructions)
    self.assertEqual(actual, {'/a/b/f1': 'f1', '/a/b/f2': 'f2'})

  def test_get_tfds_id_prefixes_multiple_directories(self):
    file_instructions = [
        shard_utils.FileInstruction('/a/b/f1', 0, -1, 1),
        shard_utils.FileInstruction('/a/b/f2', 0, -1, 1),
        shard_utils.FileInstruction('/x/y/z/f1', 0, -1, 1),
        shard_utils.FileInstruction('/a/c/f1', 0, -1, 1),
        shard_utils.FileInstruction('/a/b/c/d/e/f1', 0, -1, 1),
    ]
    actual = reader_lib._get_tfds_id_prefixes(file_instructions)
    expected = {
        '/a/b/c/d/e/f1': 'e/f1',
        '/a/b/f1': 'b/f1',
        '/a/b/f2': 'b/f2',
        '/a/c/f1': 'c/f1',
        '/x/y/z/f1': 'z/f1',
    }
    self.assertEqual(actual, expected)


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""unnatural-instructions dataset."""

from __future__ import annotations
import json
from typing import Generator, Any, Tuple, Mapping
import tensorflow_datasets.public_api as tfds

_DS_PATH = 'https://github.com/orhonovich/unnatural-instructions/raw/main/data/full_data.zip'

_DESCRIPTION = """
Dataset described in the paper: Unnatural Instructions: Tuning Language Models with (Almost) No Human Labor (2022).
Contains sets of natural-language instructions, with optional constraints / LLM-generated reformulations.
"""

_CITATION = """
@misc{honovich2022unnatural,
      title = {Unnatural Instructions: Tuning Language Models with (Almost) No Human Labor},
      author = {Honovich, Or and Scialom, Thomas and Levy, Omer and Schick, Timo},
      url = {https://arxiv.org/abs/2212.09689},
      publisher = {arXiv},
      year={2022}
}
"""


class UnnaturalInstructions(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for unnatural_instructions dataset."""

  __count__ = 0

  VERSION = tfds.core.Version('0.0.1')
  RELEASE_NOTES = {
      '0.0.1': (
          'Initial release. Omit instructions / inputs, as they require '
          'additional processing to be used. Instruction_with_inputs and '
          'reformulations contain instructions and contexts.'
      ),
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        # Per author's recommendation, we discard the ans_simple field.
        features=tfds.features.FeaturesDict({
            'id': tfds.features.Text(doc='Unique identifier for example.'),
            'instruction': tfds.features.Text(
                doc='Instruction with placeholder for inputs.'
            ),
            'instances': tfds.features.Sequence({
                'input': tfds.features.Text(
                    doc=(
                        'Input to be fed into placeholders for given'
                        ' instruction.'
                    )
                ),
                'constraints': tfds.features.Text(
                    doc='Task-specific constraints.'
                ),
                'instruction_with_input': tfds.features.Text(
                    doc='Instructions with inputs supplied to placeholders.'
                ),
                'output': tfds.features.Text(
                    doc='Target output for given task.'
                ),
            }),
            'reformulations': tfds.features.Sequence({
                'input': tfds.features.Text(
                    doc=(
                        'Input to be fed into placeholders for given'
                        ' instruction.'
                    )
                ),
                'instruction': tfds.features.Text(
                    doc='Instruction with placeholder for inputs.'
                ),
                'instruction_with_input': tfds.features.Text(
                    doc='Instructions with inputs supplied to placeholders.'
                ),
                'output': tfds.features.Text(
                    doc='Target output for given task.'
                ),
            }),
        }),
        homepage='https://github.com/orhonovich/unnatural-instructions',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(_DS_PATH)
    return {'train': self._generate_examples(path)}

  def _generate_examples(
      self,
      jsonl_path,
  ) -> Generator[Tuple[str, Mapping[str, Any]], None, None]:
    """Parses available JSON files in a given split, returns TF examples."""
    count = 0
    for file in [f.open() for f in jsonl_path.glob('*')]:
      for line in file:
        ex = json.loads(line)
        # Note: We perform minimal data massaging to preserve original struct.
        if 'reformulations' not in ex:
          ex['reformulations'] = []
        ex['id'] = str(count)
        count += 1
        yield str(count), ex  # Use count as a pseudo-id.



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrapper around FeatureDict to allow better control over decoding."""

from __future__ import annotations

import enum
import functools
from typing import Any, List, Union

from tensorflow_datasets.core import example_parser
from tensorflow_datasets.core import example_serializer
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf


class _Backend(enum.Enum):
  NP = 'numpy'
  TF = 'tensorflow'


class TopLevelFeature(feature_lib.FeatureConnector):
  """Top-level `FeatureConnector` to manage decoding.

  Note that `FeatureConnector` which are declared as `TopLevelFeature` can be
  nested. However, only the top-level feature should be decoded.

  `TopLevelFeature` allows better control over the decoding, and
  eventually better support for augmentations.
  """

  @functools.cached_property
  def flat_features(self) -> List[Any]:
    return self._flatten(self)

  @functools.cached_property
  def flat_serialized_info(self) -> List[Any]:
    return self._flatten(self.get_serialized_info())

  @functools.cached_property
  def flat_sequence_ranks(self) -> List[int]:
    return [_get_sequence_rank(s) for s in self.flat_serialized_info]  # pylint: disable=not-an-iterable

  def _decode_example_generic(
      self, serialized_example, *, decoders=None, backend=_Backend.TF
  ):
    """Decode the serialize examples.

    Args:
      serialized_example: Nested `dict` of `tf.Tensor`
      decoders: Nested dict of `Decoder` objects which allow to customize the
        decoding. The structure should match the feature structure, but only
        customized feature keys need to be present. See [the
        guide](https://github.com/tensorflow/datasets/blob/master/docs/decode.md)
        for more info.
      backend: whether to use `decode_example` (TensorFlow) or
        `decode_example_np` (NumPy).

    Returns:
      example: Nested `dict` containing the decoded nested examples.
    """
    # Step 1: Flatten the nested dict => []
    flat_example = self._flatten(serialized_example)
    flat_decoders = self._flatten(decoders)

    # Step 2: Apply the decoding
    flatten_decoded = [
        _decode_feature(  # pylint: disable=g-complex-comprehension
            feature=feature,
            example=example,
            decoder=decoder,
            sequence_rank=sequence_rank,
            backend=backend,
        )
        for (
            feature,
            example,
            decoder,
            sequence_rank,
        ) in zip(
            self.flat_features,
            flat_example,
            flat_decoders,
            self.flat_sequence_ranks,
        )
    ]

    # Step 3: Restore nesting [] => {}
    nested_decoded = self._nest(flatten_decoded)
    return nested_decoded

  def decode_example(self, serialized_example, *, decoders=None):
    return self._decode_example_generic(
        serialized_example, decoders=decoders, backend=_Backend.TF
    )

  def decode_example_np(self, serialized_example, *, decoders=None):
    return self._decode_example_generic(
        serialized_example, decoders=decoders, backend=_Backend.NP
    )

  def serialize_example(self, example_data) -> bytes:
    """Encodes nested data values into `tf.train.Example` bytes.

    See `deserialize_example` to decode the proto into `tf.Tensor`.

    Args:
      example_data: Example data to encode (numpy-like nested dict)

    Returns:
      The serialized `tf.train.Example`.
    """
    example_data = self.encode_example(example_data)
    return self._example_serializer.serialize_example(example_data)

  def deserialize_example(
      self,
      serialized_example: Union[tf.Tensor, bytes],
      *,
      decoders=None,
  ) -> utils.TensorDict:
    """Decodes the `tf.train.Example` data into `tf.Tensor`.

    See `serialize_example` to encode the data into proto.

    Args:
      serialized_example: The tensor-like object containing the serialized
        `tf.train.Example` proto.
      decoders: Eventual decoders to apply (see
        [documentation](https://www.tensorflow.org/datasets/decode))

    Returns:
      The decoded features tensors.
    """
    example_data = self._example_parser.parse_example(serialized_example)
    return self.decode_example(example_data, decoders=decoders)

  def deserialize_example_np(
      self,
      serialized_example: Union[tf.Tensor, bytes],
      *,
      decoders=None,
  ) -> utils.NpArrayOrScalarDict:
    example_data = self._example_parser_np.parse_example(serialized_example)
    return self.decode_example_np(example_data, decoders=decoders)

  @property
  def tf_example_spec(self) -> dict[str, Any]:
    """Returns the `tf.Example` proto structure.

    Returns:
      The flat `dict[str, tf.io.FixedLenFeature | tf.io.XyzFeature | ...]`
    """
    return self._example_parser.flat_feature_specs

  @functools.cached_property
  def _example_parser(self):
    example_specs = self.get_serialized_info()
    return example_parser.ExampleParser(example_specs)

  @functools.cached_property
  def _example_parser_np(self):
    example_specs = self.get_serialized_info()
    return example_parser.ExampleParserNp(example_specs)

  @functools.cached_property
  def _example_serializer(self):
    example_specs = self.get_serialized_info()
    return example_serializer.ExampleSerializer(example_specs)


def _decode_feature(
    feature,
    example,
    decoder,
    sequence_rank: int,
    backend: _Backend,
):
  """Decode a single feature."""
  if decoder is not None:
    # If the decoder is still a dict, it means that the feature is a Dataset
    # (it wasn't flattened).
    if isinstance(decoder, dict):
      decode_kwargs = dict(decoders=decoder)
      decoder = feature
    else:
      # Eventually overwrite the default decoding
      decode_kwargs = {}
      decoder.setup(feature=feature)
  else:
    decode_kwargs = {}
    decoder = feature

  if backend == _Backend.NP:
    return decoder.decode_example_np(example, **decode_kwargs)
  elif backend == _Backend.TF:
    if sequence_rank == 0:
      return decoder.decode_example(example, **decode_kwargs)
    elif sequence_rank == 1:
      # Return a batch of examples from a sequence
      return decoder.decode_batch_example(example, **decode_kwargs)
    elif sequence_rank > 1:
      # Use ragged tensor if the sequance rank is greater than one
      return decoder.decode_ragged_example(example, **decode_kwargs)
  else:
    raise ValueError(f'wrong backend {backend} to decode features.')


def _get_sequence_rank(serialized_info) -> int:
  """Return the number of sequence dimensions of the feature."""

  if isinstance(serialized_info, dict):
    # If the element is a dictionary, it might correspond to a nested dataset
    # whose serialized_info is not flattened (so it might be a nested dict).
    all_sequence_rank = [
        _get_sequence_rank(s) for s in serialized_info.values()
    ]
  else:
    # If this is a nested dataset, we ignore the sequence_rank. We will decode
    # the full dataset example with the Dataset decoder.
    if serialized_info.dataset_lvl > 0:
      return 0
    all_sequence_rank = [serialized_info.sequence_rank]
  sequence_ranks = set(all_sequence_rank)
  if len(sequence_ranks) != 1:
    raise NotImplementedError(
        'Decoding do not support mixing sequence and context features within a '
        'single FeatureConnector. Received inputs of different sequence_rank: '
        '{}'.format(sequence_ranks)
    )
  return next(iter(sequence_ranks))



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""glove_100_angular dataset."""

import h5py
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
Pre-trained Global Vectors for Word Representation (GloVe) embeddings for
approximate nearest neighbor search. This dataset consists of two splits:

  1. 'database': consists of 1,183,514 data points, each has features:
    'embedding' (100 floats), 'index' (int64), 'neighbors' (empty list).
  2. 'test': consists of 10,000 data points, each has features: 'embedding' (100
    floats), 'index' (int64), 'neighbors' (list of 'index' and 'distance'
    of the nearest neighbors in the database.)
"""

_CITATION = """
@inproceedings{pennington2014glove,
  author = {Jeffrey Pennington and Richard Socher and Christopher D. Manning},
  booktitle = {Empirical Methods in Natural Language Processing (EMNLP)},
  title = {GloVe: Global Vectors for Word Representation},
  year = {2014},
  pages = {1532--1543},
  url = {http://www.aclweb.org/anthology/D14-1162},
}
"""

_URL = 'http://ann-benchmarks.com/glove-100-angular.hdf5'


class Glove100Angular(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for glove_100_angular dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'index': tfds.features.Scalar(
                dtype=np.int64, doc='Index within the split.'
            ),
            'embedding': tfds.features.Tensor(shape=(100,), dtype=np.float32),
            'neighbors': tfds.features.Sequence(
                {
                    'index': tfds.features.Scalar(
                        dtype=np.int64, doc='Neighbor index.'
                    ),
                    'distance': tfds.features.Scalar(
                        dtype=np.float32, doc='Neighbor distance.'
                    ),
                },
                doc=(
                    'The computed neighbors, which is only available for'
                    ' the test split.'
                ),
            ),
        }),
        homepage='https://nlp.stanford.edu/projects/glove/',
        citation=_CITATION,
        disable_shuffling=True,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract({'file': _URL})
    return {
        'database': self._generate_examples(path, True),
        'test': self._generate_examples(path, False),
    }

  def _generate_examples(self, path, is_database):
    """Pulls data from hdf5 into splits."""
    with tf.io.gfile.GFile(path['file'], 'rb') as f:
      with h5py.File(f, 'r') as dataset_file:
        if is_database:
          for idx, val in enumerate(dataset_file['train']):
            yield idx, {'index': idx, 'embedding': val, 'neighbors': []}
        else:
          for idx, (val, n_indices, distances) in enumerate(
              zip(
                  dataset_file['test'],
                  dataset_file['neighbors'],
                  dataset_file['distances'],
              )
          ):
            neighbors = []
            for nn_idx, nn_dist in zip(n_indices, distances):
              neighbors.append({'index': nn_idx, 'distance': nn_dist})
            yield idx, {
                'index': idx,
                'embedding': val,
                'neighbors': neighbors,
            }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataset class for Food-101 dataset."""

import json
import os

from etils import epath
import tensorflow_datasets.public_api as tfds

_BASE_URL = "http://data.vision.ee.ethz.ch/cvl/food-101.tar.gz"

_DESCRIPTION = (
    "This dataset consists of 101 food categories, with 101'000 images. For "
    "each class, 250 manually reviewed test images are provided as well as 750"
    " training images. On purpose, the training images were not cleaned, and "
    "thus still contain some amount of noise. This comes mostly in the form of"
    " intense colors and sometimes wrong labels. All images were rescaled to "
    "have a maximum side length of 512 pixels."
)

_LABELS_FNAME = "image_classification/food-101_classes.txt"

_CITATION = """\
 @inproceedings{bossard14,
  title = {Food-101 -- Mining Discriminative Components with Random Forests},
  author = {Bossard, Lukas and Guillaumin, Matthieu and Van Gool, Luc},
  booktitle = {European Conference on Computer Vision},
  year = {2014}
}
"""


class Food101(tfds.core.GeneratorBasedBuilder):
  """Food-101 Images dataset."""

  VERSION = tfds.core.Version("2.0.0")
  SUPPORTED_VERSIONS = [
      tfds.core.Version(
          "1.0.0",
          tfds_version_to_prepare="8cea22f06d74d5848608fe7ac6d6faac7bc05b55",
      ),
      tfds.core.Version("2.1.0"),
  ]

  def _info(self):
    """Define Dataset Info."""

    names_file = tfds.core.tfds_path(_LABELS_FNAME)
    features_dict = {
        "image": tfds.features.Image(),
        "label": tfds.features.ClassLabel(names_file=names_file),
    }
    if self.version > "2.0.0":
      features_dict["id"] = tfds.features.Text()
    return tfds.core.DatasetInfo(
        builder=self,
        description=(_DESCRIPTION),
        features=tfds.features.FeaturesDict(features_dict),
        supervised_keys=("image", "label"),
        homepage="https://data.vision.ee.ethz.ch/cvl/datasets_extra/food-101/",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Define Splits."""

    dl_path = dl_manager.download_and_extract(_BASE_URL)
    meta_path = os.path.join(dl_path, "food-101", "meta")
    image_dir_path = os.path.join(dl_path, "food-101", "images")

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "json_file_path": os.path.join(meta_path, "train.json"),
                "image_dir_path": image_dir_path,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "json_file_path": os.path.join(meta_path, "test.json"),
                "image_dir_path": image_dir_path,
            },
        ),
    ]

  def _generate_examples(self, json_file_path, image_dir_path):
    """Generate images and labels for splits."""
    with epath.Path(json_file_path).open() as f:
      data = json.loads(f.read())
    for label, images in data.items():
      for image_name in images:
        image = os.path.join(image_dir_path, image_name + ".jpg")
        features = {"image": image, "label": label}
        if self.version > "2.0.0":
          features["id"] = image_name
        yield image_name, features



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Util to generate the dataset documentation content.

Used by tensorflow_datasets/scripts/documentation/build_catalog.py
"""

import collections
import dataclasses
import json
import os
import textwrap
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

import tensorflow_datasets as tfds
from tensorflow_datasets.core import constants
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

# Dict of `full_names_dict['dataset']['config']['version']`
FullNamesDict = Dict[str, Dict[str, Dict[str, Any]]]
# Same as `FullNamesDict`, but contains `True` for nightly datasets:
# * New dataset: nightly_dict['dataset'] is True
# * New config: nightly_dict['dataset']['config'] is True
# * New version: nightly_dict['dataset']['config']['version'] is True
NightlyDict = Dict[str, Union[bool, Dict[str, Union[bool, Dict[str, bool]]]]]


def get_pwc_catalog_urls() -> Mapping[str, str]:
  with open(constants.PWC_LINKS_PATH, 'r') as f:
    return json.load(f)


@dataclasses.dataclass
class DocUtilPaths:
  """Structure containing the utils paths."""

  # VisualizationDocUtil
  fig_base_path: Optional[tfds.typing.PathLike] = tfds.core.gcs_path(
      'visualization/fig/'
  )
  fig_base_url: str = (
      'https://storage.googleapis.com/tfds-data/visualization/fig/'
  )
  # DataframeDocUtil
  df_base_path: Optional[tfds.typing.PathLike] = tfds.core.gcs_path(
      'visualization/dataframe'
  )
  df_base_url: str = (
      'https://storage.googleapis.com/tfds-data/visualization/dataframe/'
  )
  # NightlyDocUtil
  nightly_path: Optional[tfds.typing.PathLike] = tfds.core.utils.tfds_path(
      'stable_versions.txt'
  )


class VisualizationDocUtil(object):
  """Small util which generate the path/urls for the visualizations."""

  def __init__(self, base_path: tfds.typing.PathLike, base_url: str):
    """Constructor.

    Args:
      base_path: Path where images are stored.
      base_url: Base url where images are displayed.
    """
    self._base_path = base_path
    self._base_url = base_url

  def _get_name(self, builder):
    return builder.info.full_name.replace('/', '-') + '.png'

  def get_url(self, builder):
    return self._base_url + self._get_name(builder)

  def get_html_tag(self, builder: tfds.core.DatasetBuilder) -> str:
    """Returns the <img> html tag."""
    url = self.get_url(builder)
    return f'<img src="{url}" alt="Visualization" width="500px">'

  def has_visualization(self, builder):
    filepath = os.path.join(self._base_path, self._get_name(builder))
    return tf.io.gfile.exists(filepath)


class DataframeDocUtil(object):
  """Small util which generates the path/urls for the dataframes."""

  def __init__(self, base_path: tfds.typing.PathLike, base_url: str):
    """Constructor.

    Args:
      base_path: Path where images are stored.
      base_url: Base url where images are displayed.
    """
    self._base_path = base_path
    self._base_url = base_url

  def _get_name(self, builder):
    return builder.info.full_name.replace('/', '-') + '.html'

  def get_url(self, builder):
    return self._base_url + self._get_name(builder)

  def get_html_tag(self, builder: tfds.core.DatasetBuilder) -> str:
    """Returns the html tag."""
    url = self.get_url(builder)
    button_id = 'displaydataframe'
    content_id = 'dataframecontent'
    visualization_html = f"""
    <!-- mdformat off(HTML should not be auto-formatted) -->

    {{% framebox %}}

    <button id="{button_id}">Display examples...</button>
    <div id="{content_id}" style="overflow-x:auto"></div>
    <script>
    const url = "{url}";
    const dataButton = document.getElementById('{button_id}');
    dataButton.addEventListener('click', async () => {{
      // Disable the button after clicking (dataframe loaded only once).
      dataButton.disabled = true;

      const contentPane = document.getElementById('{content_id}');
      try {{
        const response = await fetch(url);
        // Error response codes don't throw an error, so force an error to show
        // the error message.
        if (!response.ok) throw Error(response.statusText);

        const data = await response.text();
        contentPane.innerHTML = data;
      }} catch (e) {{
        contentPane.innerHTML =
            'Error loading examples. If the error persist, please open '
            + 'a new issue.';
      }}
    }});
    </script>

    {{% endframebox %}}

    <!-- mdformat on -->
    """
    return textwrap.dedent(visualization_html)

  def has_visualization(self, builder):
    filepath = os.path.join(self._base_path, self._get_name(builder))
    return tf.io.gfile.exists(filepath)


def _split_full_name(full_name: str) -> Tuple[str, str, str]:
  """Extracts the `(ds name, config, version)` from the full_name."""
  if not tfds.core.load.is_full_name(full_name):
    raise ValueError(
        f'Parsing builder name string {full_name} failed.'
        'The builder name string must be of the following format:'
        '`dataset_name[/config_name]/version`'
    )
  ds_name, *optional_config, version = full_name.split('/')
  assert len(optional_config) <= 1
  config = next(iter(optional_config)) if optional_config else ''
  return ds_name, config, version


def _full_names_to_dict(full_names: List[str]) -> FullNamesDict:
  """Creates the dict `d['dataset']['config']['version']`."""
  full_names_dict = collections.defaultdict(
      lambda: collections.defaultdict(  # pylint: disable=g-long-lambda
          lambda: collections.defaultdict(type(None))
      )
  )
  for full_name in full_names:
    ds_name, config, version = _split_full_name(full_name)
    full_names_dict[ds_name][config][version]  # pylint: disable=pointless-statement
  return full_names_dict


def _build_nightly_dict(
    registered_ds: FullNamesDict,
    stable_version_ds: FullNamesDict,
) -> NightlyDict:
  """Computes the nightly dict from the registered and stable dict."""
  nightly_ds = collections.defaultdict(
      lambda: collections.defaultdict(  # pylint: disable=g-long-lambda
          lambda: collections.defaultdict(bool)
      )
  )
  for dataset in registered_ds:
    if dataset in stable_version_ds:
      for config in registered_ds[dataset]:
        if config in stable_version_ds[dataset]:
          for version in registered_ds[dataset][config]:
            if version in stable_version_ds[dataset][config]:
              # (dataset, config, version) already exists
              # We add it to the nightly dict to make sure the
              # key exists
              nightly_ds[dataset][config][version] = False
            else:
              # New version only present in tfds-nightly
              nightly_ds[dataset][config][version] = True
        else:
          # New config only present in tfds-nightly
          nightly_ds[dataset][config] = True
    else:
      # New dataset only present in tfds-nightly
      nightly_ds[dataset] = True
  return nightly_ds


@tfds.core.utils.memoize()
def _load_nightly_dict(version_path: tfds.typing.PathLike) -> NightlyDict:
  """Loads (and caches) the nightly dict."""
  with tf.io.gfile.GFile(os.fspath(version_path), 'r') as f:
    stable_versions = f.read().splitlines()

  # Build the `full_names_dict['dataset']['config']['version']` for both
  # nightly and stable version
  registered_ds = _full_names_to_dict(tfds.core.load.list_full_names())
  stable_version_ds = _full_names_to_dict(stable_versions)

  # Nightly versions are `registered - stable`
  return _build_nightly_dict(registered_ds, stable_version_ds)


class NightlyDocUtil(object):
  """Small util to format the doc."""

  def __init__(self, path: tfds.typing.PathLike):
    """Constructor.

    Args:
      path: Path containing the nightly versions
    """
    self._nightly_dict: NightlyDict = _load_nightly_dict(path)

  def is_builder_nightly(
      self,
      builder: Union[tfds.core.DatasetBuilder, str],
  ) -> bool:
    """Returns `True` if the builder is new."""
    if isinstance(builder, tfds.core.DatasetBuilder):
      builder_name = builder.name
    else:
      builder_name = builder
    return self._nightly_dict[builder_name] is True  # pylint: disable=g-bool-id-comparison

  def is_config_nightly(self, builder: tfds.core.DatasetBuilder) -> bool:
    """Returns `True` if the config is new."""
    ds_name, config, _ = _split_full_name(builder.info.full_name)
    if self.is_builder_nightly(builder):
      return False
    return self._nightly_dict[ds_name][config] is True  # pylint: disable=g-bool-id-comparison

  def is_version_nightly(
      self,
      builder: tfds.core.DatasetBuilder,
      version: str,
  ) -> bool:
    """Returns `True` if the version is new."""
    ds_name, config, _ = _split_full_name(builder.info.full_name)
    if self.is_builder_nightly(builder) or self.is_config_nightly(builder):
      return False
    return self._nightly_dict[ds_name][config][version] is True  # pylint: disable=g-bool-id-comparison

  def has_nightly(self, builder: tfds.core.DatasetBuilder) -> bool:
    """Returns True if any of the builder/config/version is new."""

    def reduce(value):
      if isinstance(value, bool):
        return value
      elif isinstance(value, dict):
        return any(reduce(x) for x in value.values())
      else:
        raise AssertionError(f'Invalid nightly_dict value: {value}')

    return reduce(self._nightly_dict[builder.name])

  icon = (
      '<span class="material-icons" '
      'title="Available only in the tfds-nightly package">nights_stay</span>'
  )


def format_homepage_url(homepage):
  """Formats a URL as required for the homepage section."""
  return f'[{homepage}]({homepage})'



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Generate gtzan data.

"""

import os
import random

from absl import app
from absl import flags
from tensorflow_datasets.audio.gtzan import gtzan
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.testing import fake_data_utils

flags.DEFINE_string(
    "tfds_dir",
    os.fspath(utils.tfds_write_path()),
    "Path to tensorflow_datasets directory",
)
FLAGS = flags.FLAGS


def _output_dir():
  return os.path.join(FLAGS.tfds_dir, "audio", "gtzan", "dummy_data")


def _generate_data():
  """Generate data examples."""
  # pylint: disable=protected-access
  labels = tfds.features.ClassLabel(names=gtzan._CLASS_LABELS).names  # pytype: disable=module-attr
  # pylint: enable=protected-access
  wav_file = fake_data_utils.get_random_wav_c1(duration=30, sample=22050)
  # .wav file has name in the format of <genre>.<number>.wav
  label = random.choice(labels)
  random_number_for_filename = random.randint(0, 99)
  filename = "{}.{:05d}".format(label, random_number_for_filename)
  filepath = os.path.join(
      _output_dir(), "genres", label, "{}.wav".format(filename)
  )
  dirname = os.path.dirname(filepath)
  if not tf.io.gfile.exists(dirname):
    tf.io.gfile.makedirs(dirname)
  tf.io.gfile.copy(wav_file, filepath, overwrite=True)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")
  _generate_data()


if __name__ == "__main__":
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Splits related API."""

from __future__ import annotations

import abc
import dataclasses
import functools
import itertools
import math
import operator
import os
import re
import typing
from typing import Any, Dict, Iterable, List, Optional, Union

from absl import logging
from etils import epath
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import proto as proto_lib
from tensorflow_datasets.core import units
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils import shard_utils

from tensorflow_metadata.proto.v0 import statistics_pb2

# Accepted split values (in `as_dataset(split=...)`)
SplitArg = Union[str, 'AbstractSplit']

# <split_name>[<split_selector>] (e.g. `train[54%:]`)
_SUB_SPEC_RE = re.compile(
    r"""^
    (?P<split_name>[\w-]+)
    (\[
      (?P<split_selector>[\d\w%:-]+)
    \])?
    $""",
    re.X,  # Ignore whitespace
)
# <val><unit> (e.g. `-54%`)
_SLICE_RE = re.compile(
    r"""^
    (
        (?P<val>-?[\d_]+)
        (?P<unit>(?:%|shard))?
    )?
    $""",
    re.X,  # Ignore whitespace
)

_ADDITION_SEP_RE = re.compile(r'\s*\+\s*')


@dataclasses.dataclass(frozen=True)
class _AbsoluteInstruction:
  """A machine friendly slice: defined absolute positive boundaries."""

  splitname: str
  from_: int  # uint (starting index).
  to: int  # uint (ending index).

  def to_absolute(self, split_infos) -> List['_AbsoluteInstruction']:
    del split_infos  # unused
    return [self]


@dataclasses.dataclass(eq=False, frozen=True)
class SplitInfo:
  """Wraps `proto.SplitInfo` with an additional property.

  Attributes:
    name: Name of the split (e.g. `train`, `test`,...)
    shard_lengths: List of length <number of files> containing the number of
      examples stored in each file.
    filename_template: The template used to create sharded filenames.
    num_examples: Total number of examples (`sum(shard_lengths)`)
    num_shards: Number of files (`len(shard_lengths)`)
    num_bytes: Size of the files (in bytes)
    statistics: Additional statistics of the split.
  """

  name: str
  shard_lengths: List[int]
  num_bytes: int
  filename_template: Optional[naming.ShardedFileTemplate] = None
  statistics: statistics_pb2.DatasetFeatureStatistics = dataclasses.field(
      default_factory=statistics_pb2.DatasetFeatureStatistics,
  )

  def __post_init__(self):
    if (
        self.filename_template
        and self.filename_template.split
        and self.name != self.filename_template.split
    ):
      raise ValueError(
          f'Split name {self.name} must be equal to split name '
          f'in template {self.filename_template.split}'
      )
    if self.filename_template and not self.filename_template.split:
      super().__setattr__(
          'filename_template', self.filename_template.replace(split=self.name)
      )
    # Normalize bytes
    super().__setattr__('num_bytes', units.Size(self.num_bytes))

  @classmethod
  def from_proto(
      cls,
      proto: proto_lib.SplitInfo,
      filename_template: naming.ShardedFileTemplate,
  ) -> 'SplitInfo':
    """Returns a SplitInfo class instance from a SplitInfo proto."""
    return cls(
        name=proto.name,
        shard_lengths=list(proto.shard_lengths),
        num_bytes=proto.num_bytes,
        filename_template=filename_template.replace(
            split=proto.name, template=proto.filepath_template
        ),
        statistics=proto.statistics,
    )

  def to_proto(self) -> proto_lib.SplitInfo:
    if self.filename_template:
      filepath_template = self.filename_template.template
    else:
      filepath_template = naming.DEFAULT_FILENAME_TEMPLATE
    return proto_lib.SplitInfo(
        name=self.name,
        shard_lengths=self.shard_lengths,
        num_bytes=self.num_bytes,
        statistics=self.statistics if self.statistics.ByteSize() else None,
        filepath_template=filepath_template,
    )

  @property
  def num_examples(self) -> int:
    return sum(self.shard_lengths)

  @property
  def num_shards(self) -> int:
    return len(self.shard_lengths)

  def __repr__(self) -> str:
    num_examples = self.num_examples or 'unknown'
    return (
        f'<SplitInfo num_examples={num_examples}, num_shards={self.num_shards}>'
    )

  @property
  def file_instructions(self) -> List[shard_utils.FileInstruction]:
    """Returns the list of dict(filename, take, skip).

    This allows for creating your own `tf.data.Dataset` using the low-level
    TFDS values.

    Example:

    ```
    file_instructions = info.splits['train[75%:]'].file_instructions
    instruction_ds = tf.data.Dataset.from_generator(
        lambda: file_instructions,
        output_types={
            'filename': tf.string,
            'take': tf.int64,
            'skip': tf.int64,
        },
    )
    ds = instruction_ds.interleave(
        lambda f: tf.data.TFRecordDataset(
            f['filename']).skip(f['skip']).take(f['take'])
    )
    ```

    When `skip=0` and `take=-1`, the full shard will be read, so the `ds.skip`
    and `ds.take` could be skipped.

    Returns:
      A `dict(filename, take, skip)`
    """
    return _make_file_instructions(
        split_infos=[self],
        instruction=str(self.name),
    )

  @property
  def filenames(self) -> List[str]:
    """Returns the list of filenames."""
    if not self.filename_template:
      raise ValueError('No filename templates available.')
    return sorted(
        self.filename_template.sharded_filenames(len(self.shard_lengths))
    )

  @property
  def filepaths(self) -> List[epath.Path]:
    """All the paths for all the files that are part of this split."""
    if not self.filename_template:
      raise ValueError('No filename templates available.')
    return sorted(
        self.filename_template.sharded_filepaths(len(self.shard_lengths))
    )

  def replace(self, **kwargs: Any) -> 'SplitInfo':
    """Returns a copy of the `SplitInfo` with updated attributes."""
    return dataclasses.replace(self, **kwargs)


@dataclasses.dataclass(eq=False, frozen=True)
class MultiSplitInfo(SplitInfo):
  """SplitInfo for when a split is spread out over multiple folders.

  This should only be used to read data and not when producing data.
  """

  split_infos: List[SplitInfo] = dataclasses.field(default_factory=list)

  def __init__(self, name: str, split_infos: List[SplitInfo]):
    if not split_infos:
      raise ValueError('Need to pass a non-empty list of SplitInfos')
    object.__setattr__(self, 'split_infos', split_infos)

    # Compute attributes of parent class SplitInfo
    shard_lengths = []
    for split_info in split_infos:
      shard_lengths.extend(split_info.shard_lengths)
    # Note that num_bytes=0 means that the size isn't known. However, we do
    # compute the sum because it will provide at least some information about
    # the size, even though it may be incomplete.
    num_bytes = sum(si.num_bytes for si in split_infos)
    super().__init__(
        name=name,
        shard_lengths=shard_lengths,
        num_bytes=num_bytes,
        filename_template=None,
    )

  def to_proto(self) -> proto_lib.SplitInfo:
    # The SplitInfo proto only supports a single split.
    raise RuntimeError('to_proto is not supported on MultiSplitInfo')

  def __repr__(self) -> str:
    return (
        f'{self.__class__.__name__}('
        f'name={self.name!r}, '
        f'split_infos={self.split_infos!r})'
    )

  @property
  def file_instructions(self) -> List[shard_utils.FileInstruction]:
    result = []
    for split_info in self.split_infos:
      result.extend(split_info.file_instructions)
    return result

  @property
  def filenames(self) -> List[str]:
    """Returns the list of filenames."""
    result = []
    for split_info in self.split_infos:
      result.extend(split_info.filenames)
    return result

  @property
  def filepaths(self) -> List[epath.Path]:
    """All the paths for all the files that are part of this split."""
    result = []
    for split_info in self.split_infos:
      result.extend(split_info.filepaths)
    return result

  def replace(self, **kwargs: Any) -> 'MultiSplitInfo':
    raise RuntimeError('replace is not supported on MultiSplitInfo')


@dataclasses.dataclass(eq=False, frozen=True)
class SubSplitInfo:
  """Wrapper around a sub split info.

  This class expose info on the subsplit:

  ```
  ds, info = tfds.load(..., split='train[75%:]', with_info=True)
  info.splits['train[75%:]'].num_examples
  ```
  """

  name: str
  file_instructions: List[shard_utils.FileInstruction]

  @property
  def shard_lengths(self) -> List[int]:
    return [f.take for f in self.file_instructions]

  @property
  def num_examples(self) -> int:
    """Returns the number of example in the subsplit."""
    return sum(self.shard_lengths)

  @property
  def num_bytes(self) -> int:
    return 0  # Unknown

  @property
  def num_shards(self) -> int:
    return len(self.file_instructions)

  @property
  def filenames(self) -> List[str]:
    """Returns the list of filenames."""
    return sorted(os.path.basename(f.filename) for f in self.file_instructions)

  @property
  def filepaths(self) -> List[epath.Path]:
    """Returns the list of filepaths."""
    return sorted(epath.Path(f.filename) for f in self.file_instructions)

  def to_proto(self) -> proto_lib.SplitInfo:
    return proto_lib.SplitInfo(
        name=self.name,
        shard_lengths=self.shard_lengths,
        num_bytes=self.num_bytes,
        statistics=None,
    )


# TODO(epot): `: tfds.Split` type should be `Union[str, Split]`
class Split(str):
  # pylint: disable=line-too-long
  """`Enum` for dataset splits.

  Datasets are typically split into different subsets to be used at various
  stages of training and evaluation.

  * `TRAIN`: the training data.
  * `VALIDATION`: the validation data. If present, this is typically used as
    evaluation data while iterating on a model (e.g. changing hyperparameters,
    model architecture, etc.).
  * `TEST`: the testing data. This is the data to report metrics on. Typically
    you do not want to use this during model iteration as you may overfit to it.
  * `ALL`: All splits from the dataset merged together (`'train+test+...'`).

  See the
  [guide on
  splits](https://github.com/tensorflow/datasets/blob/master/docs/splits.md)
  for more information.
  """

  def __repr__(self) -> str:
    return '{}({})'.format(type(self).__name__, super(Split, self).__repr__())  # pytype: disable=wrong-arg-types


Split.TRAIN = Split('train')
Split.TEST = Split('test')
Split.VALIDATION = Split('validation')
Split.ALL = Split('all')

if typing.TYPE_CHECKING:
  # For type checking, `tfds.Split` is an alias for `str` with additional
  # `.TRAIN`, `.TEST`,... attributes. All strings are valid split type.
  Split = Union[Split, str]


class SplitDict(utils.NonMutableDict[str, SplitInfo]):
  """Split info object."""

  def __init__(
      self,
      split_infos: Iterable[SplitInfo],
      *,
      # TODO(b/216470058): remove this parameter
      dataset_name: Optional[str] = None,  # deprecated, please don't use
  ):
    super(SplitDict, self).__init__(
        {split_info.name: split_info for split_info in split_infos},
        error_msg='Split {key} already present',
    )
    if dataset_name:
      logging.warning(
          "DEPRECATED: SplitDict's dataset_name parameter is "
          'deprecated and can be removed.'
      )
    self._dataset_name = dataset_name  # deprecated, please don't use

  def __getitem__(self, key):
    if not self:
      raise KeyError(
          f'Trying to access `splits[{key!r}]` but `splits` is empty. '
          'This likely indicate the dataset has not been generated yet.'
      )
    # 1st case: The key exists: `info.splits['train']`
    elif str(key) in self.keys():
      return super(SplitDict, self).__getitem__(str(key))
    # 2nd case: Uses instructions: `info.splits['train[50%]']`
    else:
      instructions = _make_file_instructions(
          split_infos=list(self.values()),
          instruction=key,
      )
      return SubSplitInfo(name=key, file_instructions=instructions)

  @classmethod
  def from_proto(
      cls,
      repeated_split_infos: Iterable[proto_lib.SplitInfo],
      filename_template: naming.ShardedFileTemplate,
  ) -> 'SplitDict':
    """Returns a new SplitDict initialized from the `repeated_split_infos`."""
    split_infos = [
        SplitInfo.from_proto(
            proto=s, filename_template=filename_template.replace(split=s.name)
        )
        for s in repeated_split_infos
    ]
    return cls(split_infos)

  def to_proto(self):
    """Returns a list of SplitInfo protos that we have."""
    return [s.to_proto() for s in self.values()]

  @property
  def total_num_examples(self):
    """Return the total number of examples."""
    return sum(s.num_examples for s in self.values())

  @classmethod
  def merge_multiple(cls, split_dicts: List['SplitDict']) -> 'SplitDict':
    info_per_split = []
    for split in set(itertools.chain(*split_dicts)):
      infos_of_split = []
      for split_dict in split_dicts:
        if split not in split_dict:
          continue
        split_info = split_dict[split]
        if isinstance(split_info, MultiSplitInfo):
          infos_of_split.extend(split_info.split_infos)
        else:
          infos_of_split.append(split_info)
      info_per_split.append(
          MultiSplitInfo(name=split, split_infos=infos_of_split)
      )

    return cls(split_infos=info_per_split)


def _make_absolute_instructions(
    split_infos: Iterable[SplitInfo],
    instruction: SplitArg,
) -> List[_AbsoluteInstruction]:
  if isinstance(instruction, str):
    instruction = AbstractSplit.from_spec(instruction)

  # Create the absolute instruction (per split)
  split_info_map = {split_info.name: split_info for split_info in split_infos}
  return instruction.to_absolute(split_info_map)


def _file_instructions_for_split(
    instruction: _AbsoluteInstruction,
    split_info: SplitInfo,
) -> List[shard_utils.FileInstruction]:
  """Returns the file instructions from the given instruction applied to the given split info."""
  if not split_info.num_examples:
    raise ValueError(
        "Shard empty. This might means that dataset hasn't been generated "
        'yet and info not restored from GCS, or that legacy dataset is used.'
    )
  to = split_info.num_examples if instruction.to is None else instruction.to
  return shard_utils.get_file_instructions(
      from_=instruction.from_ or 0,
      to=to,
      filenames=[os.fspath(fp) for fp in split_info.filepaths],
      shard_lengths=split_info.shard_lengths,
  )


def _make_file_instructions(
    split_infos: List[SplitInfo],
    instruction: SplitArg,
) -> List[shard_utils.FileInstruction]:
  """Returns file instructions by applying the given instruction on the given splits.

  Args:
    split_infos: Dataset splits information
    instruction: `ReadInstruction` or `str`

  Returns:
    List of FileInstruction instances
  """

  # TODO(epot): Should try to merge the instructions together as well as
  # performing additional validation. For example, should raise an error
  # if there is overlap between splits (`train[:50]+train[:25]`)
  # If there is a single shard, `train[:25]+train[50:75]` could be optimized
  # into a single `ds.take(25).skip(50-25).take(75-50)`

  absolute_instructions = _make_absolute_instructions(
      split_infos=split_infos, instruction=instruction
  )
  instructions = []
  info_per_split = {split_info.name: split_info for split_info in split_infos}
  for abs_instr in absolute_instructions:
    split_info = info_per_split[str(abs_instr.splitname)]
    instructions.extend(
        _file_instructions_for_split(
            instruction=abs_instr, split_info=split_info
        )
    )
  return instructions


class AbstractSplit(abc.ABC):
  """Abstract base class of splits.

  Abstract splits are combined together, then passed to
  `tfds.load(..., split=)` or `builder.as_dataset(split=...)`.

  See the guide: https://www.tensorflow.org/datasets/splits
  """

  @classmethod
  def from_spec(cls, spec: SplitArg) -> 'AbstractSplit':
    """Creates a ReadInstruction instance out of a string spec.

    Args:
      spec (str): split(s) + optional slice(s) to read. A slice can be
        specified, using absolute numbers (int) or percentages (int). E.g.
        `test`: test split. `test + validation`: test split + validation split.
        `test[10:]`: test split, minus its first 10 records. `test[:10%]`: first
        10% records of test split. `test[:-5%]+train[40%:60%]`: first 95% of
        test + middle 20% of train.

    Returns:
      The split instance.
    """
    if isinstance(spec, AbstractSplit):
      return spec

    spec = str(spec)  # Need to convert to str in case of `Split` instance.

    subs = _ADDITION_SEP_RE.split(spec)
    if not subs:
      raise ValueError(f'No instructions could be built out of {spec!r}')
    with utils.try_reraise(
        f'Error parsing split {spec!r}. See format at: '
        'https://www.tensorflow.org/datasets/splits\n'
    ):
      instructions = [_str_to_relative_instruction(s) for s in subs]
    # Merge all splits together (_SplitAll)
    return functools.reduce(operator.add, instructions)

  @abc.abstractmethod
  def to_absolute(self, split_infos) -> List[_AbsoluteInstruction]:
    """Translate instruction into a list of absolute instructions.

    Those absolute instructions are then to be added together.

    Args:
      split_infos: `tfds.core.SplitDict` dict associating split names to split
        info.

    Returns:
      list of _AbsoluteInstruction instances (corresponds to the + in spec).
    """
    raise NotImplementedError

  def __add__(self, other: Union[str, 'AbstractSplit']) -> 'AbstractSplit':
    """Sum of 2 splits."""
    if not isinstance(other, (str, AbstractSplit)):
      raise TypeError(f'Adding split {self!r} with non-split value: {other!r}')
    if isinstance(other, str):  # Normalize strings
      other = AbstractSplit.from_spec(other)
    return _SplitAdd(self, other)


@dataclasses.dataclass(frozen=True)
class _SplitAdd(AbstractSplit):
  """Sum of 2 splits.

  `'train+test'` is equivalent to
  `_SplitAdd(ReadInstruction('train'), ReadInstruction('test'))`
  """

  left: AbstractSplit
  right: AbstractSplit

  def __repr__(self):
    return f'{self.left!r}+{self.right!r}'

  def to_absolute(self, split_infos) -> List[_AbsoluteInstruction]:
    # Merge instructions from left and right
    return self.left.to_absolute(split_infos) + self.right.to_absolute(
        split_infos
    )


class _SplitAll(AbstractSplit):
  """Union of all splits of the dataset."""

  def to_absolute(self, split_infos) -> List[_AbsoluteInstruction]:
    # Create the union of all splits
    split_names = split_infos.keys()
    split = AbstractSplit.from_spec('+'.join(split_names))
    return split.to_absolute(split_infos)


@dataclasses.dataclass(frozen=True)
class ReadInstruction(AbstractSplit):
  """Reading instruction for a dataset.

  See the guide: https://www.tensorflow.org/datasets/splits

  Attributes:
    split_name: name of the split to read. Eg: 'train'.
    from_: Starting index, or None if no lower boundary.
    to: Ending index, or None if no upper boundary.
    unit: optional, one of: '%': to set the slicing unit as percents of the
      split size. 'abs': to set the slicing unit as absolute numbers. 'shard':
      to set the slicing unit as shard.
    rounding: The rounding behaviour to use when percent slicing is used.
      Ignored when slicing with absolute indices. Possible values: - 'closest'
      (default): The specified percentages are rounded to the closest value. Use
      this if you want specified percents to be as much exact as possible. -
      'pct1_dropremainder': the specified percentages are treated as multiple of
      1%. Use this option if you want consistency. Eg: len(5%) == 5 * len(1%).
      Using this option, one might not be able to use the full set of examples,
      if the number of those is not a multiple of 100.
  """

  split_name: str
  # TODO(py3.10): Add `_ = dataclasses.KW_ONLY`
  from_: Optional[int] = None
  to: Optional[int] = None
  unit: str = 'abs'
  rounding: str = 'closest'

  def __post_init__(self):
    # Perform validation
    allowed_units = ['%', 'abs', 'shard']
    allowed_rounding = ['closest', 'pct1_dropremainder']
    if self.unit not in allowed_units:
      raise ValueError(
          f'Unit should be one of {allowed_units}. Got {self.unit!r}'
      )
    if self.rounding not in allowed_rounding:
      raise ValueError(
          f'Rounding should be one of {allowed_rounding}. '
          f'Got: {self.rounding!r}'
      )
    if self.unit == '%':
      if abs(self.from_ or 0) > 100 or abs(self.to or 0) > 100:
        raise ValueError(
            'When unit=%, percent slice boundaries should be '
            f'in [-100, 100]. Got: {self}'
        )

  def __repr__(self) -> str:
    unit = '' if self.unit == 'abs' else self.unit
    from_ = '' if self.from_ is None else f'{self.from_}{unit}'
    to = '' if self.to is None else f'{self.to}{unit}'
    if self.from_ is None and self.to is None:
      slice_str = ''  # Full split selected
    else:
      slice_str = f'[{from_}:{to}]'
    rounding = f', rounding={self.rounding!r}' if self.unit == '%' else ''
    return f"ReadInstruction('{self.split_name}{slice_str}'{rounding})"

  def to_absolute(self, split_infos) -> List[_AbsoluteInstruction]:
    return [_rel_to_abs_instr(self, split_infos)]


def _str_to_relative_instruction(spec: str) -> AbstractSplit:
  """Returns ReadInstruction for given string."""
  # <split_name>[<split_selector>] (e.g. `train[54%:]`)
  res = _SUB_SPEC_RE.match(spec)
  err_msg = (
      f'Unrecognized split format: {spec!r}. See format at '
      'https://www.tensorflow.org/datasets/splits'
  )
  if not res:
    raise ValueError(err_msg)
  split_name = res.group('split_name')
  split_selector = res.group('split_selector')

  if split_name == 'all':
    if split_selector:
      # TODO(tfds): `all[:75%]` could be supported by creating a
      # `_SliceSplit(split, from_=, to=, unit=)`.
      raise NotImplementedError(
          f'{split_name!r} does not support slice. Please open a github issue '
          'if you need this feature.'
      )
    return _SplitAll()

  if split_selector is None:  # split='train'
    from_ = None
    to = None
    unit = 'abs'
  else:  # split='train[x:y]' or split='train[x]'
    slices = [_SLICE_RE.match(x) for x in split_selector.split(':')]
    # Make sure all slices are valid, and at least one is not empty
    if not all(slices) or not any(x.group(0) for x in slices):  # pytype: disable=attribute-error  # re-none
      raise ValueError(err_msg)
    if len(slices) == 1:
      (from_match,) = slices
      from_ = from_match['val']
      to = int(from_) + 1
      unit = from_match['unit'] or 'abs'
      if unit != 'shard':
        raise ValueError('Absolute or percent only support slice syntax.')
    elif len(slices) == 2:
      from_match, to_match = slices
      from_ = from_match['val']
      to = to_match['val']
      unit = from_match['unit'] or to_match['unit'] or 'abs'
    else:
      raise ValueError(err_msg)

  if from_ is not None:
    from_ = int(from_)
  if to is not None:
    to = int(to)

  return ReadInstruction(
      split_name=split_name,
      rounding='closest',
      from_=from_,
      to=to,
      unit=unit,
  )


def _pct_to_abs_pct1(boundary, num_examples: int):
  # Using math.trunc here, since -99.5% should give -99%, not -100%.
  if num_examples < 100:
    msg = (
        'Using "pct1_dropremainder" rounding on a split with less than 100 '
        'elements is forbidden: it always results in an empty dataset.'
    )
    raise ValueError(msg)
  return boundary * math.trunc(num_examples / 100.0)


def _pct_to_abs_closest(boundary, num_examples: int) -> int:
  return int(round(boundary * num_examples / 100.0))


def _rel_to_abs_instr(
    rel_instr: ReadInstruction,
    split_infos: Dict[str, SplitInfo],
) -> _AbsoluteInstruction:
  """Returns _AbsoluteInstruction instance for given RelativeInstruction.

  Args:
    rel_instr: ReadInstruction instance.
    split_infos: dict {split_name: split_infos}.
  """
  pct_to_abs = (
      _pct_to_abs_closest
      if rel_instr.rounding == 'closest'
      else _pct_to_abs_pct1
  )
  split = rel_instr.split_name
  if split not in split_infos:
    raise ValueError(
        f'Unknown split {split!r}. Should be one of {list(split_infos)}.'
    )
  num_examples = split_infos[split].num_examples
  from_ = rel_instr.from_
  to = rel_instr.to
  if rel_instr.unit == '%':
    from_ = 0 if from_ is None else pct_to_abs(from_, num_examples)
    to = num_examples if to is None else pct_to_abs(to, num_examples)
  elif rel_instr.unit == 'shard':
    shard_lengths = split_infos[split].shard_lengths
    from_ = 0 if from_ is None else sum(shard_lengths[:from_])
    if to is not None and to <= 0:
      to = len(shard_lengths) + to
    to = num_examples if to is None else sum(shard_lengths[:to])
  elif rel_instr.unit == 'abs':
    from_ = 0 if from_ is None else from_
    to = num_examples if to is None else to
  else:
    raise ValueError(f'Invalid split unit: {rel_instr.unit}')
  if abs(from_) > num_examples or abs(to) > num_examples:
    msg = 'Requested slice [%s:%s] incompatible with %s examples.' % (
        from_ or '',
        to or '',
        num_examples,
    )
    raise ValueError(msg)
  if from_ < 0:
    from_ = num_examples + from_
  elif from_ == 0:
    from_ = None
  if to < 0:
    to = num_examples + to
  elif to == num_examples:
    to = None
  return _AbsoluteInstruction(split, from_, to)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""mslr_web dataset."""

import dataclasses
import itertools

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.ranking.libsvm_ranking_parser import LibSVMRankingParser

_DESCRIPTION = """
MSLR-WEB are two large-scale Learning-to-Rank datasets released by Microsoft
Research. The first dataset (called "30k") contains 30,000 queries and the
second dataset (called "10k") contains 10,000 queries. Each dataset consists of
query-document pairs represented as feature vectors and corresponding relevance
judgment labels.

You can specify whether to use the "10k" or "30k" version of the dataset, and a
corresponding fold, as follows:

```python
ds = tfds.load("mslr_web/30k_fold1")
```

If only `mslr_web` is specified, the `mslr_web/10k_fold1` option is selected by
default:

```python
# This is the same as `tfds.load("mslr_web/10k_fold1")`
ds = tfds.load("mslr_web")
```
"""

_CITATION = """
@article{DBLP:journals/corr/QinL13,
  author    = {Tao Qin and Tie{-}Yan Liu},
  title     = {Introducing {LETOR} 4.0 Datasets},
  journal   = {CoRR},
  volume    = {abs/1306.2597},
  year      = {2013},
  url       = {http://arxiv.org/abs/1306.2597},
  timestamp = {Mon, 01 Jul 2013 20:31:25 +0200},
  biburl    = {http://dblp.uni-trier.de/rec/bib/journals/corr/QinL13},
  bibsource = {dblp computer science bibliography, http://dblp.org}
}
"""

_URLS = {
    "10k": "https://api.onedrive.com/v1.0/shares/s!AtsMfWUz5l8nbOIoJ6Ks0bEMp78/root/content",
    "30k": "https://api.onedrive.com/v1.0/shares/s!AtsMfWUz5l8nbXGPBlwD1rnFdBY/root/content",
}

_FEATURE_NAMES = {
    1: "covered_query_term_number_body",
    2: "covered_query_term_number_anchor",
    3: "covered_query_term_number_title",
    4: "covered_query_term_number_url",
    5: "covered_query_term_number_whole_document",
    6: "covered_query_term_ratio_body",
    7: "covered_query_term_ratio_anchor",
    8: "covered_query_term_ratio_title",
    9: "covered_query_term_ratio_url",
    10: "covered_query_term_ratio_whole_document",
    11: "stream_length_body",
    12: "stream_length_anchor",
    13: "stream_length_title",
    14: "stream_length_url",
    15: "stream_length_whole_document",
    16: "idf_body",
    17: "idf_anchor",
    18: "idf_title",
    19: "idf_url",
    20: "idf_whole_document",
    21: "sum_of_term_frequency_body",
    22: "sum_of_term_frequency_anchor",
    23: "sum_of_term_frequency_title",
    24: "sum_of_term_frequency_url",
    25: "sum_of_term_frequency_whole_document",
    26: "min_of_term_frequency_body",
    27: "min_of_term_frequency_anchor",
    28: "min_of_term_frequency_title",
    29: "min_of_term_frequency_url",
    30: "min_of_term_frequency_whole_document",
    31: "max_of_term_frequency_body",
    32: "max_of_term_frequency_anchor",
    33: "max_of_term_frequency_title",
    34: "max_of_term_frequency_url",
    35: "max_of_term_frequency_whole_document",
    36: "mean_of_term_frequency_body",
    37: "mean_of_term_frequency_anchor",
    38: "mean_of_term_frequency_title",
    39: "mean_of_term_frequency_url",
    40: "mean_of_term_frequency_whole_document",
    41: "variance_of_term_frequency_body",
    42: "variance_of_term_frequency_anchor",
    43: "variance_of_term_frequency_title",
    44: "variance_of_term_frequency_url",
    45: "variance_of_term_frequency_whole_document",
    46: "sum_of_stream_length_normalized_term_frequency_body",
    47: "sum_of_stream_length_normalized_term_frequency_anchor",
    48: "sum_of_stream_length_normalized_term_frequency_title",
    49: "sum_of_stream_length_normalized_term_frequency_url",
    50: "sum_of_stream_length_normalized_term_frequency_whole_document",
    51: "min_of_stream_length_normalized_term_frequency_body",
    52: "min_of_stream_length_normalized_term_frequency_anchor",
    53: "min_of_stream_length_normalized_term_frequency_title",
    54: "min_of_stream_length_normalized_term_frequency_url",
    55: "min_of_stream_length_normalized_term_frequency_whole_document",
    56: "max_of_stream_length_normalized_term_frequency_body",
    57: "max_of_stream_length_normalized_term_frequency_anchor",
    58: "max_of_stream_length_normalized_term_frequency_title",
    59: "max_of_stream_length_normalized_term_frequency_url",
    60: "max_of_stream_length_normalized_term_frequency_whole_document",
    61: "mean_of_stream_length_normalized_term_frequency_body",
    62: "mean_of_stream_length_normalized_term_frequency_anchor",
    63: "mean_of_stream_length_normalized_term_frequency_title",
    64: "mean_of_stream_length_normalized_term_frequency_url",
    65: "mean_of_stream_length_normalized_term_frequency_whole_document",
    66: "variance_of_stream_length_normalized_term_frequency_body",
    67: "variance_of_stream_length_normalized_term_frequency_anchor",
    68: "variance_of_stream_length_normalized_term_frequency_title",
    69: "variance_of_stream_length_normalized_term_frequency_url",
    70: "variance_of_stream_length_normalized_term_frequency_whole_document",
    71: "sum_of_tf_idf_body",
    72: "sum_of_tf_idf_anchor",
    73: "sum_of_tf_idf_title",
    74: "sum_of_tf_idf_url",
    75: "sum_of_tf_idf_whole_document",
    76: "min_of_tf_idf_body",
    77: "min_of_tf_idf_anchor",
    78: "min_of_tf_idf_title",
    79: "min_of_tf_idf_url",
    80: "min_of_tf_idf_whole_document",
    81: "max_of_tf_idf_body",
    82: "max_of_tf_idf_anchor",
    83: "max_of_tf_idf_title",
    84: "max_of_tf_idf_url",
    85: "max_of_tf_idf_whole_document",
    86: "mean_of_tf_idf_body",
    87: "mean_of_tf_idf_anchor",
    88: "mean_of_tf_idf_title",
    89: "mean_of_tf_idf_url",
    90: "mean_of_tf_idf_whole_document",
    91: "variance_of_tf_idf_body",
    92: "variance_of_tf_idf_anchor",
    93: "variance_of_tf_idf_title",
    94: "variance_of_tf_idf_url",
    95: "variance_of_tf_idf_whole_document",
    96: "boolean_model_body",
    97: "boolean_model_anchor",
    98: "boolean_model_title",
    99: "boolean_model_url",
    100: "boolean_model_whole_document",
    101: "vector_space_model_body",
    102: "vector_space_model_anchor",
    103: "vector_space_model_title",
    104: "vector_space_model_url",
    105: "vector_space_model_whole_document",
    106: "bm25_body",
    107: "bm25_anchor",
    108: "bm25_title",
    109: "bm25_url",
    110: "bm25_whole_document",
    111: "lmir_abs_body",
    112: "lmir_abs_anchor",
    113: "lmir_abs_title",
    114: "lmir_abs_url",
    115: "lmir_abs_whole_document",
    116: "lmir_dir_body",
    117: "lmir_dir_anchor",
    118: "lmir_dir_title",
    119: "lmir_dir_url",
    120: "lmir_dir_whole_document",
    121: "lmir_jm_body",
    122: "lmir_jm_anchor",
    123: "lmir_jm_title",
    124: "lmir_jm_url",
    125: "lmir_jm_whole_document",
    126: "number_of_slash_in_url",
    127: "length_of_url",
    128: "inlink_number",
    129: "outlink_number",
    130: "page_rank",
    131: "site_rank",
    132: "quality_score",
    133: "quality_score_2",
    134: "query_url_click_count",
    135: "url_click_count",
    136: "url_dwell_time",
}

_LABEL_NAME = "label"


@dataclasses.dataclass
class MslrWebConfig(tfds.core.BuilderConfig):
  size: str = "10k"  # one of "10k" or "30k"
  fold: int = 1


class MslrWeb(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for mslr_web dataset."""

  VERSION = tfds.core.Version("1.2.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
      "1.1.0": "Bundle features into a single 'float_features' feature.",
      "1.2.0": "Add query and document identifiers.",
  }
  # pytype: disable=wrong-keyword-args
  BUILDER_CONFIGS = [
      MslrWebConfig(name=f"{size}_fold{fold}", size=size, fold=fold)
      for size, fold in itertools.product(["10k", "30k"], [1, 2, 3, 4, 5])
  ]

  # pytype: enable=wrong-keyword-args

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    encoding = tfds.features.Encoding.ZLIB
    features = {
        "float_features": tfds.features.Tensor(
            shape=(None, len(_FEATURE_NAMES)),
            dtype=np.float64,
            encoding=encoding,
        ),
        _LABEL_NAME: tfds.features.Tensor(
            shape=(None,), dtype=np.float64, encoding=encoding
        ),
        "query_id": tfds.features.Text(),
        "doc_id": tfds.features.Tensor(
            shape=(None,), dtype=np.int64, encoding=encoding
        ),
    }
    metadata = tfds.core.MetadataDict(
        {
            "float_features_names": {
                idx: _FEATURE_NAMES[key]
                for idx, key in enumerate(sorted(_FEATURE_NAMES))
            }
        }
    )

    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(features),
        homepage="https://www.microsoft.com/en-us/research/project/mslr/",
        citation=_CITATION,
        metadata=metadata,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    size = self.builder_config.size
    fold = self.builder_config.fold
    path = dl_manager.download_and_extract(_URLS[size])

    return {
        split: self._generate_examples(path / f"Fold{fold}/{split}.txt")
        for split in ["train", "vali", "test"]
    }

  def _generate_examples(self, path):
    """Yields examples."""
    with tf.io.gfile.GFile(path, "r") as f:
      yield from LibSVMRankingParser(
          f, _FEATURE_NAMES, _LABEL_NAME, combine_features=True
      )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility library to generate dataset-like files."""

import random
import tempfile

import numpy as np
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

MIN_HEIGHT_WIDTH = 10
MAX_HEIGHT_WIDTH = 15
CHANNELS_NB = 3

_SAMPLE_RATE = 44100
_AUDIO_DURATION = 5


def get_random_picture(height=None, width=None, channels=CHANNELS_NB):
  """Returns random picture as np.ndarray (int)."""
  height = height or random.randrange(MIN_HEIGHT_WIDTH, MAX_HEIGHT_WIDTH)
  width = width or random.randrange(MIN_HEIGHT_WIDTH, MAX_HEIGHT_WIDTH)
  return np.random.randint(256, size=(height, width, channels), dtype=np.uint8)


def get_random_jpeg(height=None, width=None, channels=CHANNELS_NB):
  """Returns path to JPEG picture."""
  image = get_random_picture(height, width, channels)
  jpeg = tf.image.encode_jpeg(image)
  with utils.nogpu_session() as sess:
    res = sess.run(jpeg)
  fobj = tempfile.NamedTemporaryFile(delete=False, mode='wb', suffix='.JPEG')
  fobj.write(res)
  fobj.close()
  return fobj.name


def get_random_png(height=None, width=None, channels=CHANNELS_NB):
  """Returns path to PNG picture."""
  # Big randomly generated pngs take large amounts of diskspace.
  # Instead, we resize a 4x4 random image to the png size.
  image = get_random_picture(4, 4, channels)
  if (height is not None) and (width is not None):
    image = tf.image.resize_nearest_neighbor(
        tf.expand_dims(image, 0), (height, width)
    )[0]
  png = tf.image.encode_png(image)
  with utils.nogpu_session() as sess:
    res = sess.run(png)
  fobj = tempfile.NamedTemporaryFile(delete=False, mode='wb', suffix='.PNG')
  fobj.write(res)
  fobj.close()
  return fobj.name


def get_random_audio(duration=_AUDIO_DURATION, sample=_SAMPLE_RATE):
  """Returns random audio as np.ndarray (float32)."""
  sample_number = np.arange(duration * sample)
  waveform = np.sin(2 * np.pi * sample_number * 440.0 / sample).astype(
      np.float32
  )
  waveform = waveform * 0.3
  return waveform


def get_random_wav_c1(
    channels=1, duration=_AUDIO_DURATION, sample=_SAMPLE_RATE
):
  """Returns path to WAV audio having channels = 1."""
  audio = get_random_audio(duration, sample).reshape(-1, channels)
  wav = tf.audio.encode_wav(audio, sample)
  with utils.nogpu_session() as sess:
    res = sess.run(wav)
  with tempfile.NamedTemporaryFile(delete=False, mode='wb', suffix='.wav') as f:
    f.write(res)
  return f.name


def get_random_wav_c2(
    channels=2, duration=_AUDIO_DURATION, sample=_SAMPLE_RATE
):
  """Returns path to WAV audio having channels = 2."""
  audio = get_random_audio(duration, sample).reshape(-1, channels)
  wav = tf.audio.encode_wav(audio, sample)
  with utils.nogpu_session() as sess:
    res = sess.run(wav)
  with tempfile.NamedTemporaryFile(delete=False, mode='wb', suffix='.wav') as f:
    f.write(res)
  return f.name



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for wiki40b dataset module."""

import os

from tensorflow_datasets import testing
from tensorflow_datasets.text import wiki40b

_EXAMPLE_DIR = os.path.join(
    os.path.normpath(os.path.dirname(__file__) + "/../"),
    "testing",
    "test_data",
    "fake_examples",
    "wiki40b",
)


class Wiki40bTest(testing.DatasetBuilderTestCase):

  @classmethod
  def setUpClass(cls):
    super(Wiki40bTest, cls).setUpClass()
    wiki40b._DATA_DIRECTORY = _EXAMPLE_DIR

  DATASET_CLASS = wiki40b.Wiki40b
  BUILDER_CONFIG_NAMES_TO_TEST = ["en"]

  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""wiki_bio dataset."""
from __future__ import annotations

import os

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """
@inproceedings{lebret-etal-2016-neural,
    title = "Neural Text Generation from Structured Data with Application to the Biography Domain",
    author = "Lebret, R{\'e}mi  and
      Grangier, David  and
      Auli, Michael",
    booktitle = "Proceedings of the 2016 Conference on Empirical Methods in Natural Language Processing",
    month = nov,
    year = "2016",
    address = "Austin, Texas",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/D16-1128",
    doi = "10.18653/v1/D16-1128",
    pages = "1203--1213",
}
"""

_DESCRIPTION = """
WikiBio is constructed using Wikipedia biography pages, it contains the first
paragraph and the infobox tokenized.
The dataset follows a standarized table format.
"""

_URL = 'https://huggingface.co/datasets/wiki_bio/resolve/main/data/wikipedia-biography-dataset.zip'


def _get_table(infobox_line):
  """Converts the infobox into a one row table."""
  cells = infobox_line.split('\t')
  # remove empty cells
  cells = list(filter(lambda x: x.find('<none>') == -1, cells))
  columns = set([cell[0 : cell.split(':')[0].rfind('_')] for cell in cells])
  table = {col: dict() for col in columns}
  for cell in cells:
    delimiter_position_value = cell.find(':')
    column_index = cell[0:delimiter_position_value]
    value = cell[delimiter_position_value + 1 :]
    delimiter_column_index = column_index.rfind('_')
    column = column_index[0:delimiter_column_index]
    index = column_index[delimiter_column_index + 1 :]
    table[column][index] = value
  infobox_line_as_table = []
  for column in table.keys():
    row_value = ' '.join(
        [table[column][index] for index in sorted(table[column].keys())]
    )
    infobox_line_as_table.append({
        'column_header': column,
        'row_number': 1,
        'content': row_value,
    })
  return infobox_line_as_table


class WikiBio(tfds.core.GeneratorBasedBuilder):
  """Infoboxes and first paragraph from Wikipedia biography pages."""

  VERSION = tfds.core.Version('0.1.0')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'input_text': {
                'table': tfds.features.Sequence({
                    'column_header': np.str_,
                    'row_number': np.int16,
                    'content': np.str_,
                }),
                # context will be the article's title
                'context': np.str_,
            },
            'target_text': np.str_,
        }),
        supervised_keys=('input_text', 'target_text'),
        homepage='https://github.com/DavidGrangier/wikipedia-biography-dataset',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    dataset_url = tfds.download.Resource(
        url=_URL, extract_method=tfds.download.ExtractMethod.ZIP
    )
    data_dir = dl_manager.download_and_extract(dataset_url)

    extracted_path = os.path.join(data_dir, 'wikipedia-biography-dataset')

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                'id_file': os.path.join(extracted_path, 'train', 'train.id'),
                'infobox_file': os.path.join(
                    extracted_path, 'train', 'train.box'
                ),
                'nb_lines_file': os.path.join(
                    extracted_path, 'train', 'train.nb'
                ),
                'sentences_file': os.path.join(
                    extracted_path, 'train', 'train.sent'
                ),
                'article_title_file': os.path.join(
                    extracted_path, 'train', 'train.title'
                ),
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                'id_file': os.path.join(extracted_path, 'valid', 'valid.id'),
                'infobox_file': os.path.join(
                    extracted_path, 'valid', 'valid.box'
                ),
                'nb_lines_file': os.path.join(
                    extracted_path, 'valid', 'valid.nb'
                ),
                'sentences_file': os.path.join(
                    extracted_path, 'valid', 'valid.sent'
                ),
                'article_title_file': os.path.join(
                    extracted_path, 'valid', 'valid.title'
                ),
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                'id_file': os.path.join(extracted_path, 'test', 'test.id'),
                'infobox_file': os.path.join(
                    extracted_path, 'test', 'test.box'
                ),
                'nb_lines_file': os.path.join(
                    extracted_path, 'test', 'test.nb'
                ),
                'sentences_file': os.path.join(
                    extracted_path, 'test', 'test.sent'
                ),
                'article_title_file': os.path.join(
                    extracted_path, 'test', 'test.title'
                ),
            },
        ),
    ]

  def _generate_examples(
      self,
      id_file,
      infobox_file,
      nb_lines_file,
      sentences_file,
      article_title_file,
  ):
    """Yields examples."""
    with tf.io.gfile.GFile(id_file) as id_src, tf.io.gfile.GFile(
        infobox_file
    ) as infobox_src, tf.io.gfile.GFile(
        nb_lines_file
    ) as nb_lines_src, tf.io.gfile.GFile(
        sentences_file
    ) as sentences_src, tf.io.gfile.GFile(
        article_title_file
    ) as article_title_src:
      for id_, infobox, nb_lines, article_title in zip(
          id_src, infobox_src, nb_lines_src, article_title_src
      ):
        target_text = []
        for _ in range(int(nb_lines)):
          target_text.append(sentences_src.readline())
        yield id_, {
            'input_text': {
                'table': _get_table(infobox),
                'context': article_title,
            },
            'target_text': ''.join(target_text),
        }
